#! /usr/bin/env python
"""Curated metadata project"""
from pathlib import Path

import zope.interface
from pydantic import Field

import colrev.env.utils
import colrev.ops.search
import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import Fields

# pylint: disable=too-few-public-methods
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.package_manager.interfaces.ReviewTypeInterface)
class CuratedMasterdata:
    """Curated masterdata"""

    settings_class = colrev.package_manager.package_settings.DefaultSettings
    ci_supported: bool = Field(default=True)

    def __init__(
        self, *, operation: colrev.process.operation.Operation, settings: dict
    ) -> None:
        self.settings = self.settings_class(**settings)
        self.review_manager = operation.review_manager

    def __str__(self) -> str:
        return "curated masterdata repository"

    def initialize(
        self, settings: colrev.settings.Settings
    ) -> colrev.settings.Settings:
        """Initialize a curated masterdata repository"""

        # replace readme
        colrev.env.utils.retrieve_package_file(
            template_file=Path("packages/review_types/curated_masterdata/readme.md"),
            target=Path("readme.md"),
        )
        colrev.env.utils.retrieve_package_file(
            template_file=Path(
                "packages/review_types/curated_masterdata/curations_github_colrev_update.yml"
            ),
            target=Path(".github/workflows/colrev_update.yml"),
        )

        if hasattr(self.review_manager.settings.project, "curation_url"):
            colrev.env.utils.inplace_change(
                filename=Path("readme.md"),
                old_string="{{url}}",
                new_string=self.review_manager.settings.project.curation_url,
            )

        settings.search.retrieve_forthcoming = False

        settings.prep.prep_rounds[0].prep_package_endpoints = [
            {"endpoint": "colrev.source_specific_prep"},
            {"endpoint": "colrev.exclude_complementary_materials"},
            {"endpoint": "colrev.remove_urls_with_500_errors"},
            {"endpoint": "colrev.remove_broken_ids"},
            {"endpoint": "colrev.global_ids_consistency_check"},
            {"endpoint": "colrev.get_doi_from_urls"},
            {"endpoint": "colrev.get_year_from_vol_iss_jour"},
        ]

        settings.prep.prep_man_package_endpoints = [
            {"endpoint": "colrev.prep_man_curation_jupyter"},
            {"endpoint": "colrev.export_man_prep"},
        ]
        settings.prescreen.explanation = (
            "All records are automatically prescreen included."
        )

        settings.screen.explanation = (
            "All records are automatically included in the screen."
        )

        settings.prescreen.prescreen_package_endpoints = [
            {
                "endpoint": "colrev.scope_prescreen",
                "ExcludeComplementaryMaterials": True,
            },
            {"endpoint": "colrev.conditional_prescreen"},
        ]
        settings.screen.screen_package_endpoints = []
        settings.pdf_get.pdf_get_package_endpoints = []

        settings.dedupe.dedupe_package_endpoints = [
            {
                "endpoint": "colrev.curation_full_outlet_dedupe",
                "selected_source": "data/search/CROSSREF.bib",
            },
            {
                "endpoint": "colrev.curation_full_outlet_dedupe",
                "selected_source": "data/search/pdfs.bib",
            },
            {"endpoint": "colrev.curation_missing_dedupe"},
        ]

        settings.data.data_package_endpoints = [
            {
                "endpoint": "colrev.colrev_curation",
                "version": "0.1",
                "curation_url": "TODO",
                "curated_masterdata": True,
                "masterdata_restrictions": {
                    # "1900": {
                    #     Fields.ENTRYTYPE: "article",
                    #     Fields.VOLUME: True,
                    #     Fields.NUMBER: True,
                    #     Fields.JOURNAL: "Journal Name",
                    # }
                },
                "curated_fields": [Fields.DOI, Fields.URL],
            }
        ]

        # curated repo: automatically prescreen/screen-include papers
        # (no data endpoint -> automatically rev_synthesized)

        return settings



#! /usr/bin/env python
"""Indexing and retrieving records locally."""
from __future__ import annotations

import sqlite3
import typing
from copy import deepcopy
from multiprocessing import Lock
from pathlib import Path

import git
from git.exc import GitCommandError

import colrev.env.environment_manager
import colrev.env.local_index_sqlite
import colrev.env.resources
import colrev.env.tei_parser
import colrev.env.utils
import colrev.exceptions as colrev_exceptions
import colrev.loader.load_utils
import colrev.ops.check
import colrev.record.record
from colrev.constants import ENTRYTYPES
from colrev.constants import Fields
from colrev.env.local_index_prep import prepare_record_for_return


class LocalIndex:
    """The LocalIndex implements indexing and retrieval of records across projects"""

    def __init__(
        self,
        *,
        index_tei: bool = False,
        verbose_mode: bool = False,
    ) -> None:
        self.verbose_mode = verbose_mode
        self.environment_manager = colrev.env.environment_manager.EnvironmentManager()
        self._index_tei = index_tei
        self.thread_lock = Lock()

    def get_journal_rankings(self, journal: str) -> list:
        """Get the journal rankings from the sqlite database"""
        sqlite_index_ranking = colrev.env.local_index_sqlite.SQLiteIndexRankings()
        return sqlite_index_ranking.select(journal=journal)

    def _retrieve_based_on_colrev_id(
        self, cids_to_retrieve: list
    ) -> colrev.record.record.Record:

        sqlite_index_record = colrev.env.local_index_sqlite.SQLiteIndexRecord()
        for cid_to_retrieve in cids_to_retrieve:
            try:
                retrieved_record = sqlite_index_record.get(
                    key=Fields.COLREV_ID, value=cid_to_retrieve
                )
                return colrev.record.record.Record(retrieved_record)

            except colrev_exceptions.RecordNotInIndexException:
                continue  # continue with the next cid_to_retrieve
            finally:
                sqlite_index_record.connection.close()

        raise colrev_exceptions.RecordNotInIndexException(cids_to_retrieve[0])

    def _retrieve_from_github_curation(
        self, record_dict: dict
    ) -> colrev.record.record.Record:  # pragma: no cover
        ret = {}
        try:
            gh_url, record_id = record_dict[Fields.CURATION_ID].split("#")
            temp_path = Path.home().joinpath("colrev").joinpath("test")
            temp_path.mkdir(exist_ok=True, parents=True)
            target_path = Path(temp_path) / Path(gh_url.split("/")[-1])
            if not target_path.is_dir():
                git.Repo.clone_from(
                    gh_url,  # .replace("https://github.com/", "git@github.com:") + ".git",
                    str(target_path),
                    depth=1,
                )
            ret = colrev.loader.load_utils.load(
                filename=target_path / Path("data/records.bib"),
            )

        except GitCommandError:
            pass

        if record_id not in ret:
            raise colrev_exceptions.RecordNotInIndexException(record_dict[Fields.ID])

        ret[record_id][Fields.CURATION_ID] = record_dict[Fields.CURATION_ID]
        return colrev.record.record.Record(ret[record_id])

    def _retrieve_from_record_index(
        self, record_dict: dict
    ) -> colrev.record.record.Record:

        record = colrev.record.record.Record(record_dict)
        cids_to_retrieve = [record.get_colrev_id()]
        retrieved_record = self._retrieve_based_on_colrev_id(cids_to_retrieve)
        if retrieved_record.data[Fields.ENTRYTYPE] != record.data[Fields.ENTRYTYPE]:
            if record_dict.get(Fields.CURATION_ID, "NA").startswith(
                "https://github.com/"
            ):
                return self._retrieve_from_github_curation(record_dict=record_dict)
            raise colrev_exceptions.RecordNotInIndexException(record_dict[Fields.ID])
        return retrieved_record

    def search(self, query: str) -> list[colrev.record.record.Record]:
        """Run a search for records"""

        try:
            self.thread_lock.acquire(timeout=60)
            sqlite_index_record = colrev.env.local_index_sqlite.SQLiteIndexRecord()
            records_to_return = []
            for record_dict in sqlite_index_record.search(query=query):
                record = prepare_record_for_return(record_dict, include_file=False)
                records_to_return.append(record)

        except sqlite3.OperationalError as exc:  # pragma: no cover
            print(exc)
        finally:
            sqlite_index_record.connection.close()
            self.thread_lock.release()

        return records_to_return

    def get_year_from_toc(self, record_dict: dict) -> str:
        """Determine the year of a paper based on its table-of-content (journal-volume-number)"""

        try:
            sqlite_index_toc = colrev.env.local_index_sqlite.SQLiteIndexTOC()
            toc_key = colrev.record.record.Record(record_dict).get_toc_key()
            toc_items = []
            if self._toc_exists(toc_key):
                toc_items = sqlite_index_toc.get_toc_items(toc_key=toc_key)

            if not toc_items:
                raise colrev_exceptions.TOCNotAvailableException()

            toc_records_colrev_id = toc_items[0]
            sqlite_index_record = colrev.env.local_index_sqlite.SQLiteIndexRecord()
            record_dict = sqlite_index_record.get(
                key=Fields.COLREV_ID, value=toc_records_colrev_id
            )

            year = record_dict.get(Fields.YEAR, "NA")
            return year

        except (
            colrev_exceptions.NotEnoughDataToIdentifyException,
            colrev_exceptions.NotTOCIdentifiableException,
            colrev_exceptions.RecordNotInIndexException,
        ) as exc:
            raise colrev_exceptions.TOCNotAvailableException() from exc
        finally:
            sqlite_index_toc.connection.close()

    def _toc_exists(self, toc_item: str) -> bool:
        try:
            self.thread_lock.acquire(timeout=60)
            sqlite_index_toc = colrev.env.local_index_sqlite.SQLiteIndexTOC()
            return sqlite_index_toc.exists(toc_item)
        except sqlite3.OperationalError:  # pragma: no cover
            pass  # return False
        except AttributeError:  # pragma: no cover
            # ie. no sqlite database available
            pass  # return False
        finally:
            if sqlite_index_toc is not None:  # Check if it was initialized
                sqlite_index_toc.connection.close()
            self.thread_lock.release()
        return False

    def _get_toc_items(self, toc_key: str, *, search_across_tocs: bool) -> list:
        sqlite_index_toc = colrev.env.local_index_sqlite.SQLiteIndexTOC()
        toc_items = []
        if self._toc_exists(toc_key):
            toc_items = sqlite_index_toc.get_toc_items(toc_key=toc_key)
        else:
            if not search_across_tocs:
                sqlite_index_toc.connection.close()
                raise colrev_exceptions.RecordNotInIndexException(toc_key)

        if not toc_items and search_across_tocs:
            try:

                partial_toc_key = toc_key.rsplit("|", 1)[0]

                toc_items = sqlite_index_toc.get_toc_items(
                    partial_toc_key=partial_toc_key
                )
                sqlite_index_toc.connection.close()
            except (
                colrev_exceptions.NotTOCIdentifiableException,
                KeyError,
            ) as exc:
                raise colrev_exceptions.RecordNotInIndexException(
                    partial_toc_key
                ) from exc

        if not toc_items:
            raise colrev_exceptions.RecordNotInIndexException(toc_key)
        return toc_items

    def retrieve_from_toc(
        self,
        record: colrev.record.record.Record,
        *,
        include_file: bool = False,
        search_across_tocs: bool = False,
    ) -> colrev.record.record.Record:
        """Retrieve a record from the toc (table-of-contents)"""

        # Note: in NotTOCIdentifiableException cases, we still need a toc_key.
        # to accomplish this, the get_toc_key() may acced an "accept_incomplete" flag
        try:
            toc_key = record.get_toc_key()
        except colrev_exceptions.NotTOCIdentifiableException as exc:
            raise colrev_exceptions.RecordNotInIndexException(
                record.data[Fields.ID]
            ) from exc

        toc_items = self._get_toc_items(toc_key, search_across_tocs=search_across_tocs)
        # SQLiteIndexRecord() must be after _get_toc_items(), which also uses the sqlite file
        sqlite_index_record = colrev.env.local_index_sqlite.SQLiteIndexRecord()
        try:
            for toc_records_colrev_id in toc_items:
                record_dict = sqlite_index_record.get(
                    key=Fields.COLREV_ID, value=toc_records_colrev_id
                )

                if not colrev.record.record_similarity.matches(
                    record, colrev.record.record.Record(record_dict)
                ):
                    continue

                return prepare_record_for_return(record_dict, include_file=include_file)
            raise colrev_exceptions.RecordNotInTOCException(
                record_id=record.data[Fields.ID], toc_key=toc_key
            )

        except (
            colrev_exceptions.NotEnoughDataToIdentifyException,
            colrev_exceptions.NotTOCIdentifiableException,
        ):
            pass

        sqlite_index_record.connection.close()

        raise colrev_exceptions.RecordNotInIndexException(record.data[Fields.ID])

    def retrieve_based_on_colrev_pdf_id(
        self, *, colrev_pdf_id: str
    ) -> colrev.record.record.Record:
        """
        Convenience function to retrieve the indexed record_dict metadata
        based on a colrev_pdf_id
        """
        try:
            sqlite_index_record = colrev.env.local_index_sqlite.SQLiteIndexRecord()
            record_dict = sqlite_index_record.get(
                key=Fields.PDF_ID, value=colrev_pdf_id
            )
            record_to_import = prepare_record_for_return(record_dict, include_file=True)
            record_to_import.data.pop(Fields.FILE, None)
        finally:
            sqlite_index_record.connection.close()
        return record_to_import

    def retrieve(
        self,
        record_dict: dict,
        *,
        include_file: bool = False,
        include_colrev_ids: bool = False,
    ) -> colrev.record.record.Record:
        """
        Convenience function to retrieve the indexed record_dict metadata
        based on another record_dict
        """

        # To avoid modifications to the original record
        record_dict = deepcopy(record_dict)

        # 1. Try the record index
        try:
            retrieved_record = self._retrieve_from_record_index(record_dict)
            retrieved_record_dict = retrieved_record.data
        except (
            colrev_exceptions.RecordNotInIndexException,
            colrev_exceptions.NotEnoughDataToIdentifyException,
        ) as exc:

            # 2. Try using global-ids
            retrieved_record_dict = {}
            remove_colrev_id = False
            if Fields.COLREV_ID not in record_dict:
                try:
                    record_dict[Fields.COLREV_ID] = colrev.record.record.Record(
                        record_dict
                    ).get_colrev_id()
                    remove_colrev_id = True
                except colrev_exceptions.NotEnoughDataToIdentifyException:
                    pass
            for key, value in record_dict.items():
                if (
                    key
                    not in colrev.env.local_index_sqlite.SQLiteIndexRecord.GLOBAL_KEYS
                    or Fields.ID == key
                ):
                    continue
                try:
                    sqlite_index_record = (
                        colrev.env.local_index_sqlite.SQLiteIndexRecord()
                    )
                    retrieved_record_dict = sqlite_index_record.get(
                        key=key, value=value
                    )
                finally:
                    sqlite_index_record.connection.close()

                if key in retrieved_record_dict:
                    if retrieved_record_dict[key] == value:
                        break
                retrieved_record_dict = {}
            if remove_colrev_id:
                del record_dict[Fields.COLREV_ID]

            if not retrieved_record_dict:
                raise colrev_exceptions.RecordNotInIndexException(
                    record_dict.get(Fields.ID, "no-key")
                ) from exc

        return prepare_record_for_return(
            retrieved_record_dict,
            include_file=include_file,
            include_colrev_ids=include_colrev_ids,
        )

    def get_fields_to_remove(self, record_dict: dict) -> list:
        """Compares the record to available toc items and
        returns fields to remove (if any), such as the volume or number."""
        # pylint: disable=too-many-return-statements

        fields_to_remove: typing.List[str] = []
        if (
            Fields.JOURNAL not in record_dict
            and record_dict[Fields.ENTRYTYPE] != ENTRYTYPES.ARTICLE
        ):
            return fields_to_remove

        internal_record_dict = deepcopy(record_dict)

        if all(
            x in internal_record_dict.keys() for x in [Fields.VOLUME, Fields.NUMBER]
        ):
            try:
                toc_key_full = colrev.record.record.Record(
                    internal_record_dict
                ).get_toc_key()

                if self._toc_exists(toc_key_full):
                    return fields_to_remove
            except colrev_exceptions.NotTOCIdentifiableException:
                return fields_to_remove
            wo_nr = deepcopy(internal_record_dict)
            del wo_nr[Fields.NUMBER]
            toc_key_wo_nr = colrev.record.record.Record(wo_nr).get_toc_key()

            if toc_key_wo_nr != "NA":
                if self._toc_exists(toc_key_wo_nr):
                    fields_to_remove.append(Fields.NUMBER)
                    return fields_to_remove

            wo_vol = deepcopy(internal_record_dict)
            del wo_vol[Fields.VOLUME]
            toc_key_wo_vol = colrev.record.record.Record(wo_vol).get_toc_key()
            if toc_key_wo_vol != "NA":
                if self._toc_exists(toc_key_wo_vol):
                    fields_to_remove.append(Fields.VOLUME)
                    return fields_to_remove

            wo_vol_nr = deepcopy(internal_record_dict)
            del wo_vol_nr[Fields.VOLUME]
            del wo_vol_nr[Fields.NUMBER]
            toc_key_wo_vol_nr = colrev.record.record.Record(wo_vol_nr).get_toc_key()
            if toc_key_wo_vol_nr != "NA":
                if self._toc_exists(toc_key_wo_vol_nr):
                    fields_to_remove.append(Fields.NUMBER)
                    fields_to_remove.append(Fields.VOLUME)
                    return fields_to_remove

        return fields_to_remove



#! /usr/bin/env python
"""Checker for erroneous-symbol-in-field."""
from __future__ import annotations

import colrev.record.qm.quality_model
from colrev.constants import DefectCodes
from colrev.constants import Fields

# pylint: disable=too-few-public-methods


class ErroneousSymbolInFieldChecker:
    """The ErroneousSymbolInFieldChecker"""

    fields_to_check = [
        Fields.AUTHOR,
        Fields.TITLE,
        Fields.EDITOR,
        Fields.JOURNAL,
        Fields.BOOKTITLE,
    ]
    erroneous_symbols = ["�", "™"]
    msg = DefectCodes.ERRONEOUS_SYMBOL_IN_FIELD

    def __init__(
        self, quality_model: colrev.record.qm.quality_model.QualityModel
    ) -> None:
        self.quality_model = quality_model

    def run(self, *, record: colrev.record.record.Record) -> None:
        """Run the erroneous-symbol-in-field checks"""

        for key in self.fields_to_check:
            if key not in record.data or record.ignored_defect(
                key=key, defect=self.msg
            ):
                continue
            if record.masterdata_is_curated():
                continue

            if any(x in record.data[key] for x in self.erroneous_symbols):
                record.add_field_provenance_note(key=key, note=self.msg)
            else:
                record.remove_field_provenance_note(key=key, note=self.msg)


def register(quality_model: colrev.record.qm.quality_model.QualityModel) -> None:
    """Register the checker"""
    quality_model.register_checker(ErroneousSymbolInFieldChecker(quality_model))



#! /usr/bin/env python
"""Convenience functions to write bib files"""
from __future__ import annotations

from copy import deepcopy
from pathlib import Path

from colrev.constants import Fields


RECORDS_FIELD_ORDER = [
    Fields.ORIGIN,  # must be in second line
    Fields.STATUS,
    Fields.MD_PROV,
    Fields.D_PROV,
    Fields.PDF_ID,
    Fields.SCREENING_CRITERIA,
    Fields.FILE,  # Note : do not change this order (parsers rely on it)
    Fields.PRESCREEN_EXCLUSION,
    Fields.DOI,
    Fields.GROBID_VERSION,
    Fields.DBLP_KEY,
    Fields.SEMANTIC_SCHOLAR_ID,
    Fields.WEB_OF_SCIENCE_ID,
    Fields.AUTHOR,
    Fields.BOOKTITLE,
    Fields.JOURNAL,
    Fields.TITLE,
    Fields.YEAR,
    Fields.VOLUME,
    Fields.NUMBER,
    Fields.PAGES,
    Fields.EDITOR,
    Fields.PUBLISHER,
    Fields.URL,
    Fields.ABSTRACT,
]


def _save_field_dict(*, input_dict: dict, input_key: str) -> list:
    list_to_return = []
    assert input_key in [Fields.MD_PROV, Fields.D_PROV]
    if input_key == Fields.MD_PROV:
        for key, value in input_dict.items():
            if isinstance(value, dict):
                formated_node = ",".join(
                    sorted(e for e in value["note"].split(",") if "" != e)
                )
                list_to_return.append(f"{key}:{value['source']};{formated_node};")

    elif input_key == Fields.D_PROV:
        for key, value in input_dict.items():
            if isinstance(value, dict):
                list_to_return.append(f"{key}:{value['source']};{value['note']};")

    return list_to_return


def _get_stringified_record(*, record_dict: dict) -> dict:
    data_copy = deepcopy(record_dict)

    def list_to_str(*, val: list) -> str:
        return ("\n" + " " * 36).join([f.rstrip() for f in val])

    for key in [Fields.ORIGIN]:
        if key in data_copy:
            if key in [Fields.ORIGIN]:
                data_copy[key] = sorted(list(set(data_copy[key])))
            for ind, val in enumerate(data_copy[key]):
                if len(val) > 0:
                    if val[-1] != ";":
                        data_copy[key][ind] = val + ";"
            data_copy[key] = list_to_str(val=data_copy[key])

    for key in [Fields.MD_PROV, Fields.D_PROV]:
        if key in data_copy:
            if isinstance(data_copy[key], dict):
                data_copy[key] = _save_field_dict(
                    input_dict=data_copy[key], input_key=key
                )
            if isinstance(data_copy[key], list):
                data_copy[key] = list_to_str(val=data_copy[key])

    return data_copy


def to_string(*, records_dict: dict) -> str:
    """Convert a records dict to a bibtex string"""
    # Note: we need a deepcopy because the parsing modifies dicts
    recs_dict = deepcopy(records_dict)

    def format_field(field: str, value: str) -> str:
        padd = " " * max(0, 28 - len(field))
        return f",\n   {field} {padd} = {{{value}}}"

    bibtex_str = ""
    first = True
    for record_id, record_dict in sorted(recs_dict.items()):
        if not first:
            bibtex_str += "\n"
        first = False

        bibtex_str += f"@{record_dict[Fields.ENTRYTYPE]}{{{record_id}"

        record_dict = _get_stringified_record(record_dict=record_dict)

        for ordered_field in RECORDS_FIELD_ORDER:
            if ordered_field in record_dict:
                if record_dict[ordered_field] == "":
                    continue
                bibtex_str += format_field(ordered_field, record_dict[ordered_field])

        for key in sorted(record_dict.keys()):
            if key in RECORDS_FIELD_ORDER + [Fields.ID, Fields.ENTRYTYPE]:
                continue

            bibtex_str += format_field(key, record_dict[key])

        bibtex_str += ",\n}\n"

    return bibtex_str


def write_file(*, records_dict: dict, filename: Path) -> None:
    """Write a bib file from a records dict"""
    bibtexstr = to_string(records_dict=records_dict)
    with open(filename, "w", encoding="utf-8") as file:
        file.write(bibtexstr)



#!/usr/bin/env python
"""Utility to transform items from arXiv into records"""
from __future__ import annotations

from colrev.constants import Fields

# pylint: disable=too-many-branches
# pylint: disable=colrev-missed-constant-usage

FIELDS_TO_REMOVE = [
    "links",
    "href",
    "guidislink",
    "summary_detail",
    "title_detail",
    "updated",
    "updated_parsed",
    "author_detail",
    "published_parsed",
    "arxiv_primary_category",
    "tags",
]


def parse_record(entry: dict) -> dict:
    """Transform an arXiv item into a record"""

    entry[Fields.ENTRYTYPE] = "techreport"
    entry["arxivid"] = entry.pop("id").replace("http://arxiv.org/abs/", "")
    entry[Fields.AUTHOR] = " and ".join([a["name"] for a in entry.pop("authors")])
    entry[Fields.YEAR] = entry.pop("published")[:4]
    entry[Fields.ABSTRACT] = entry.pop("summary")
    entry[Fields.ABSTRACT] = (
        entry[Fields.ABSTRACT].replace("\n", " ").replace("\r", " ")
    )
    if "arxiv_journal_ref" in entry:
        entry["arxiv_journal_ref"] = (
            entry["arxiv_journal_ref"].replace("\n", " ").replace("\r", " ")
        )
    entry[Fields.TITLE] = entry["title"].replace("\n ", "")
    if "arxiv_doi" in entry:
        entry[Fields.DOI] = entry.pop("arxiv_doi")
    if "links" in entry:
        for link in entry["links"]:
            if link["type"] == "application/pdf":
                entry[Fields.FULLTEXT] = link["href"]
            else:
                entry[Fields.URL] = link["href"]
    if "link" in entry:
        if "url" in entry:
            del entry["link"]
        else:
            entry[Fields.URL] = entry.pop("link")
    if "arxiv_comment" in entry:
        entry[Fields.KEYWORDS] = (
            entry.pop("arxiv_comment").replace("Key words: ", "").replace("\n", "")
        )

    for field_to_remove in FIELDS_TO_REMOVE:
        if field_to_remove in entry:
            del entry[field_to_remove]

    if "keywords" in entry and "pages" in entry["keywords"]:
        del entry["keywords"]
    return entry



#!/usr/bin/env python3
"""Settings of the CoLRev project."""
from __future__ import annotations

import json
import typing
from pathlib import Path

from pydantic import BaseModel
from pydantic import Field
from pydantic import model_validator

import colrev.env.utils
import colrev.exceptions as colrev_exceptions
import colrev.ops.search_api_feed
from colrev.constants import IDPattern
from colrev.constants import PDFPathType
from colrev.constants import ScreenCriterionType
from colrev.constants import SearchType
from colrev.constants import ShareStatReq

# Note : to avoid performance issues on startup (ReviewManager, parsing settings)
# the settings dataclasses should be in one file (13s compared to 0.3s)


# Project


class Author(BaseModel):
    """Author of the review"""

    # pylint: disable=too-many-instance-attributes

    name: str
    initials: str
    email: str
    orcid: typing.Optional[str] = None
    contributions: typing.List[str] = Field(default_factory=list)
    affiliations: typing.Optional[str] = None
    funding: typing.List[str] = Field(default_factory=list)
    identifiers: typing.List[str] = Field(default_factory=list)


class Protocol(BaseModel):
    """Review protocol"""

    url: str


class ProjectSettings(BaseModel):
    """Project settings"""

    # pylint: disable=too-many-instance-attributes

    title: str
    __doc_title__ = "The title of the review"
    authors: typing.List[Author]
    keywords: typing.List[str]
    # status ? (development/published?)
    protocol: typing.Optional[Protocol]
    # publication: ... (reference, link, ....)
    review_type: str
    id_pattern: IDPattern
    share_stat_req: ShareStatReq
    delay_automated_processing: bool
    colrev_version: str
    auto_upgrade: bool

    def __str__(self) -> str:
        project_str = f"- review ({self.review_type}):"
        project_str += f"\n- title: {self.title}"
        return project_str


# Search


class SearchSource(BaseModel):
    """Search source settings"""

    # pylint: disable=too-many-instance-attributes
    endpoint: str
    filename: Path
    search_type: SearchType
    search_parameters: dict
    comment: typing.Optional[str]

    to_import: int = 0
    imported_origins: typing.List[str] = []
    len_before: int = 0
    source_records_list: typing.List[typing.Dict] = []

    # pylint: disable=no-self-argument
    @model_validator(mode="before")
    def validate_filename(cls, values):  # type: ignore
        """Validate the filename"""
        filename = values.get("filename")
        if filename and not str(filename).replace("\\", "/").startswith("data/search"):
            raise colrev_exceptions.InvalidSettingsError(
                msg=f"Source filename does not start with data/search: {filename}"
            )
        return values

    def model_dump(self, **kwargs) -> dict:  # type: ignore
        exclude = set(kwargs.pop("exclude", set()))
        exclude.add("to_import")
        exclude.add("imported_origins")
        exclude.add("len_before")
        exclude.add("source_records_list")
        return super().model_dump(exclude=exclude, **kwargs)

    def setup_for_load(
        self,
        *,
        source_records_list: typing.List[typing.Dict],
        imported_origins: typing.List[str],
    ) -> None:
        """Set the SearchSource up for the load process (initialize statistics)"""
        # pylint: disable=attribute-defined-outside-init
        # Note : define outside init because the following
        # attributes are temporary. They should not be
        # saved to SETTINGS_FILE.

        self.to_import = len(source_records_list)
        self.imported_origins: typing.List[str] = imported_origins
        self.len_before = len(imported_origins)
        self.source_records_list: typing.List[typing.Dict] = source_records_list

    def get_origin_prefix(self) -> str:
        """Get the corresponding origin prefix"""
        assert not any(x in str(self.filename.name) for x in [";", "/"])
        return str(self.filename.name).lstrip("/")

    def is_md_source(self) -> bool:
        """Check whether the source is a metadata source (for preparation)"""

        return str(self.filename.name).startswith("md_")

    def is_curated_source(self) -> bool:
        """Check whether the source is a curated source (for preparation)"""

        return self.get_origin_prefix() == "md_curated.bib"

    def get_query(self) -> str:
        """Get the query filepath"""
        assert self.search_type == SearchType.DB
        # Note : save API queries in SETTINGS_FILE
        if "query_file" not in self.search_parameters:
            raise KeyError
        if not Path(self.search_parameters["query_file"]).is_file():
            raise FileNotFoundError
        return Path(self.search_parameters["query_file"]).read_text(encoding="utf-8")

    def get_api_feed(
        self,
        review_manager: colrev.review_manager.ReviewManager,
        source_identifier: str,
        update_only: bool,
        prep_mode: bool = False,
    ) -> colrev.ops.search_api_feed.SearchAPIFeed:
        """Get a feed to add and update records"""

        return colrev.ops.search_api_feed.SearchAPIFeed(
            review_manager=review_manager,
            source_identifier=source_identifier,
            search_source=self,
            update_only=update_only,
            prep_mode=prep_mode,
        )

    def __str__(self) -> str:
        formatted_str = (
            f"{str(self.search_type).lower()}: {self.endpoint} >> {self.filename}"
        )
        if self.search_parameters:
            formatted_str += f"\n   search parameters:   {self.search_parameters}"
        if self.comment:
            formatted_str += f"\n   comment:             {self.comment}"

        return formatted_str


class SearchSettings(BaseModel):
    """Search settings"""

    retrieve_forthcoming: bool

    def __str__(self) -> str:
        return f"- retrieve_forthcoming: {self.retrieve_forthcoming}"


# Prep


class PrepRound(BaseModel):
    """Prep round settings"""

    name: str
    prep_package_endpoints: list

    def __str__(self) -> str:
        short_list = [script["endpoint"] for script in self.prep_package_endpoints][:3]
        if len(self.prep_package_endpoints) > 3:
            short_list.append("...")
        return f"{self.name} (" + ",".join(short_list) + ")"


class PrepSettings(BaseModel):
    """Prep settings"""

    fields_to_keep: typing.List[str]
    prep_rounds: typing.List[PrepRound]

    prep_man_package_endpoints: list

    defects_to_ignore: list

    def __str__(self) -> str:
        return (
            f"- fields_to_keep: {self.fields_to_keep}\n"
            + "- prep_rounds:\n - endpoints:\n   - "
            + "\n   - ".join([str(prep_round) for prep_round in self.prep_rounds])
        )


# Dedupe


class DedupeSettings(BaseModel):
    """Dedupe settings"""

    dedupe_package_endpoints: list

    def __str__(self) -> str:
        endpoints_str = "- endpoints: []\n"
        if self.dedupe_package_endpoints:
            endpoints_str = "- endpoints:\n - " + "\n - ".join(
                [s["endpoint"] for s in self.dedupe_package_endpoints]
            )
        return endpoints_str


# Prescreen


class PrescreenSettings(BaseModel):
    """Prescreen settings"""

    explanation: str
    prescreen_package_endpoints: list

    def __str__(self) -> str:
        endpoints_str = "- endpoints: []\n"
        if self.prescreen_package_endpoints:
            endpoints_str = "- endpoints:\n - " + "\n - ".join(
                [s["endpoint"] for s in self.prescreen_package_endpoints]
            )
        return endpoints_str


# PDF get


class PDFGetSettings(BaseModel):
    """PDF get settings"""

    pdf_path_type: PDFPathType
    pdf_required_for_screen_and_synthesis: bool
    """With the pdf_required_for_screen_and_synthesis flag, the PDF retrieval
    can be specified as mandatory (true) or optional (false) for the following steps"""
    rename_pdfs: bool
    pdf_get_package_endpoints: list

    pdf_get_man_package_endpoints: list

    defects_to_ignore: list

    def __str__(self) -> str:
        endpoints_str = "- endpoints: []\n"
        if self.pdf_get_man_package_endpoints:
            endpoints_str = "- endpoints:\n - " + "\n - ".join(
                [s["endpoint"] for s in self.pdf_get_package_endpoints]
            )
        return f"- pdf_path_type: {self.pdf_path_type.value}\n" + endpoints_str


# PDF prep


class PDFPrepSettings(BaseModel):
    """PDF prep settings"""

    keep_backup_of_pdfs: bool

    pdf_prep_package_endpoints: list

    pdf_prep_man_package_endpoints: list

    def __str__(self) -> str:
        endpoints_str = "- endpoints: []\n"
        if self.pdf_prep_package_endpoints:
            endpoints_str = "- endpoints:\n - " + "\n - ".join(
                [s["endpoint"] for s in self.pdf_prep_package_endpoints]
            )
        return endpoints_str


# Screen


class ScreenCriterion(BaseModel):
    """Screen criterion"""

    explanation: str
    comment: typing.Optional[str]
    criterion_type: ScreenCriterionType

    def __str__(self) -> str:
        return f"{self.explanation} ({self.criterion_type}, {self.comment})"


class ScreenSettings(BaseModel):
    """Screen settings"""

    explanation: typing.Optional[str] = None
    criteria: typing.Dict[str, ScreenCriterion]
    screen_package_endpoints: list

    def __str__(self) -> str:
        endpoints_str = "- endpoints: []\n"
        if self.screen_package_endpoints:
            endpoints_str = "- endpoints:\n - " + "\n - ".join(
                [s["endpoint"] for s in self.screen_package_endpoints]
            )
        criteria_str = "- Criteria: []"
        if self.criteria:
            criteria_str = "- Criteria:\n - " + "\n - ".join(
                [
                    f"{short_name}: " + str(criterion)
                    for short_name, criterion in self.criteria.items()
                ]
            )
        return criteria_str + "\n" + endpoints_str


# Data


class DataSettings(BaseModel):
    """Data settings"""

    data_package_endpoints: list

    def __str__(self) -> str:
        endpoints_str = "- endpoints: []\n"
        if self.data_package_endpoints:
            endpoints_str = "- endpoints:\n - " + "\n - ".join(
                [s["endpoint"] for s in self.data_package_endpoints]
            )
        return endpoints_str


class Settings(BaseModel):
    """CoLRev project settings"""

    # pylint: disable=too-many-instance-attributes

    project: ProjectSettings
    sources: typing.List[SearchSource]
    search: SearchSettings
    prep: PrepSettings
    dedupe: DedupeSettings
    prescreen: PrescreenSettings
    pdf_get: PDFGetSettings
    pdf_prep: PDFPrepSettings
    screen: ScreenSettings
    data: DataSettings

    def model_dump(self, **kwargs) -> dict:  # type: ignore
        """Dump the settings model with recursive handling of SearchSource."""

        sources_dump = [source.model_dump(**kwargs) for source in self.sources]
        data = super().model_dump(**kwargs)
        data["sources"] = sources_dump
        return data

    def is_curated_repo(self) -> bool:
        """Check whether data is curated in this repository"""

        curation_endpoints = [
            x
            for x in self.data.data_package_endpoints
            if x["endpoint"] == "colrev.colrev_curation"
        ]
        return bool(curation_endpoints)

    def is_curated_masterdata_repo(self) -> bool:
        """Check whether the masterdata is curated in this repository"""

        curation_endpoints = [
            x
            for x in self.data.data_package_endpoints
            if x["endpoint"] == "colrev.colrev_curation"
        ]

        if curation_endpoints:
            curation_endpoint = curation_endpoints[0]
            if curation_endpoint["curated_masterdata"]:
                return True
        return False

    def __str__(self) -> str:
        sources_str = "\n- " + "\n- ".join(
            [str(s) for s in self.sources if s.is_md_source()]
            + [str(s) for s in self.sources if not s.is_md_source()]
        )

        return (
            str(self.project)
            + "\nSearch\n"
            + str(self.search)
            + "\nSources"
            + sources_str
            + "\nPreparation\n"
            + str(self.prep)
            + "\nDedupe\n"
            + str(self.dedupe)
            + "\nPrescreen\n"
            + str(self.prescreen)
            + "\nPDF get\n"
            + str(self.pdf_get)
            + "\nPDF prep\n"
            + str(self.pdf_prep)
            + "\nScreen\n"
            + str(self.screen)
            + "\nData\n"
            + str(self.data)
        )

    def get_packages(self) -> typing.List[str]:
        """Get the list of all package names"""

        def extract_endpoints(package_endpoints: list) -> list:
            return [
                e for pe in package_endpoints for k, e in pe.items() if k == "endpoint"
            ]

        all_packages = (
            [self.project.review_type]
            + [s.endpoint for s in self.sources]
            + extract_endpoints(self.prep.prep_man_package_endpoints)
            + extract_endpoints(self.dedupe.dedupe_package_endpoints)
            + extract_endpoints(self.prescreen.prescreen_package_endpoints)
            + extract_endpoints(self.pdf_get.pdf_get_package_endpoints)
            + extract_endpoints(self.pdf_get.pdf_get_man_package_endpoints)
            + extract_endpoints(self.pdf_prep.pdf_prep_package_endpoints)
            + extract_endpoints(self.pdf_prep.pdf_prep_man_package_endpoints)
            + extract_endpoints(self.screen.screen_package_endpoints)
            + extract_endpoints(self.data.data_package_endpoints)
        )
        for p_round in self.prep.prep_rounds:
            all_packages.extend(extract_endpoints(p_round.prep_package_endpoints))

        return all_packages


def _add_missing_attributes(loaded_dict: dict) -> None:  # pragma: no cover
    # replace dict with defaults if values are missing (to avoid exceptions)
    if "defects_to_ignore" not in loaded_dict["pdf_get"]:
        loaded_dict["pdf_get"]["defects_to_ignore"] = []


def _load_settings_from_dict(loaded_dict: dict) -> Settings:
    try:
        _add_missing_attributes(loaded_dict)
        settings = Settings(**loaded_dict)
        filenames = [x.filename for x in settings.sources]
        if not len(filenames) == len(set(filenames)):
            non_unique = list({str(x) for x in filenames if filenames.count(x) > 1})
            msg = f"Non-unique source filename(s): {', '.join(non_unique)}"
            raise colrev_exceptions.InvalidSettingsError(msg=msg, fix_per_upgrade=False)

    except (Exception,) as exc:  # pragma: no cover
        raise colrev_exceptions.InvalidSettingsError(msg=str(exc)) from exc

    return settings


def load_settings(*, settings_path: Path) -> Settings:
    """Load the settings from file"""

    if not settings_path.is_file():
        raise colrev_exceptions.RepoSetupError()

    try:
        with open(settings_path, encoding="utf-8") as file:
            loaded_dict = json.load(file)

    except json.decoder.JSONDecodeError as exc:
        raise colrev_exceptions.RepoSetupError(
            f"Failed to load settings: {exc}"
        ) from exc

    return _load_settings_from_dict(loaded_dict)


def save_settings(*, review_manager: colrev.review_manager.ReviewManager) -> None:
    """Save the settings"""

    exported_dict = review_manager.settings.model_dump()
    exported_dict = colrev.env.utils.custom_asdict_factory(exported_dict)

    with open(review_manager.paths.settings, "w", encoding="utf-8") as outfile:
        json.dump(exported_dict, outfile, indent=4)
    review_manager.dataset.add_changes(review_manager.paths.settings)



#! /usr/bin/env python
"""SearchSource: IEEEXplore"""
from __future__ import annotations

import typing
from pathlib import Path

import pandas as pd
import zope.interface
from pydantic import Field

import colrev.ops.prep
import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.packages.ieee.src.ieee_api
import colrev.record.record
import colrev.record.record_prep
from colrev.constants import ENTRYTYPES
from colrev.constants import Fields
from colrev.constants import SearchSourceHeuristicStatus
from colrev.constants import SearchType

# pylint: disable=unused-argument
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.package_manager.interfaces.SearchSourceInterface)
class IEEEXploreSearchSource:
    """IEEEXplore"""

    flag = True

    settings_class = colrev.package_manager.package_settings.DefaultSourceSettings

    # pylint: disable=colrev-missed-constant-usage
    source_identifier = "ID"
    search_types = [SearchType.API]
    endpoint = "colrev.ieee"

    ci_supported: bool = Field(default=True)
    heuristic_status = SearchSourceHeuristicStatus.oni

    db_url = "https://ieeexplore.ieee.org/Xplore/home.jsp"
    SETTINGS = {
        "api_key": "packages.search_source.colrev.ieee.api_key",
    }

    def __init__(
        self,
        *,
        source_operation: colrev.process.operation.Operation,
        settings: typing.Optional[dict] = None,
    ) -> None:
        self.review_manager = source_operation.review_manager

        if settings:
            self.search_source = self.settings_class(**settings)
        else:
            self.search_source = colrev.settings.SearchSource(
                endpoint=self.endpoint,
                filename=Path("data/search/ieee.bib"),
                search_type=SearchType.OTHER,
                search_parameters={},
                comment="",
            )
        self.source_operation = source_operation

    # For run_search, a Python SDK would be available:
    # https://developer.ieee.org/Python_Software_Development_Kit

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for IEEEXplore"""

        result = {"confidence": 0.1}

        if "Date Added To Xplore" in data:
            result["confidence"] = 0.9
            return result

        return result

    @classmethod
    def add_endpoint(
        cls, operation: colrev.ops.search.Search, params: str
    ) -> colrev.settings.SearchSource:
        """Add SearchSource as an endpoint (based on query provided to colrev search --add )"""

        params_dict = {}
        if params:
            if params.startswith("http"):
                params_dict = {Fields.URL: params}
            else:
                for item in params.split(";"):
                    key, value = item.split("=")
                    params_dict[key] = value

        search_type = operation.select_search_type(
            search_types=cls.search_types, params=params_dict
        )

        if search_type == SearchType.API:
            if len(params_dict) == 0:
                search_source = operation.create_api_source(endpoint=cls.endpoint)

            # pylint: disable=colrev-missed-constant-usage
            elif (
                "https://ieeexploreapi.ieee.org/api/v1/search/articles?"
                in params_dict["url"]
            ):
                query = (
                    params_dict["url"]
                    .replace(
                        "https://ieeexploreapi.ieee.org/api/v1/search/articles?", ""
                    )
                    .lstrip("&")
                )

                parameter_pairs = query.split("&")
                search_parameters = {}
                for parameter in parameter_pairs:
                    key, value = parameter.split("=")
                    search_parameters[key] = value

                last_value = list(search_parameters.values())[-1]

                filename = operation.get_unique_filename(
                    file_path_string=f"ieee_{last_value}"
                )

                search_source = colrev.settings.SearchSource(
                    endpoint=cls.endpoint,
                    filename=filename,
                    search_type=SearchType.API,
                    search_parameters=search_parameters,
                    comment="",
                )
            else:
                raise NotImplementedError

        elif search_type == SearchType.DB:
            search_source = operation.create_db_source(
                search_source_cls=cls,
                params=params_dict,
            )
        else:
            raise NotImplementedError

        operation.add_source_and_search(search_source)
        return search_source

    def search(self, rerun: bool) -> None:
        """Run a search of IEEEXplore"""

        if self.search_source.search_type == SearchType.API:
            ieee_feed = self.search_source.get_api_feed(
                review_manager=self.review_manager,
                source_identifier=self.source_identifier,
                update_only=(not rerun),
            )
            self._run_api_search(ieee_feed=ieee_feed, rerun=rerun)

        elif self.search_source.search_type == SearchType.DB:
            self.source_operation.run_db_search(  # type: ignore
                search_source_cls=self.__class__,
                source=self.search_source,
            )

        else:
            raise NotImplementedError

    def _get_api_key(self) -> str:
        api_key = self.review_manager.environment_manager.get_settings_by_key(
            self.SETTINGS["api_key"]
        )
        if api_key is None or len(api_key) != 24:
            api_key = input("Please enter api key: ")
            self.review_manager.environment_manager.update_registry(
                self.SETTINGS["api_key"], api_key
            )
        return api_key

    def _run_api_search(
        self, ieee_feed: colrev.ops.search_api_feed.SearchAPIFeed, rerun: bool
    ) -> None:
        api_key = self._get_api_key()

        api = colrev.packages.ieee.src.ieee_api.XPLORE(
            parameters=self.search_source.search_parameters, api_key=api_key
        )
        while True:
            retrieved_records = api.get_records()
            if not retrieved_records:
                break

            for retrieved_record in retrieved_records:
                ieee_feed.add_update_record(retrieved_record)

            api.startRecord += api.resultSetMax

        ieee_feed.save()

    def prep_link_md(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.record.Record:
        """Not implemented"""
        return record

    def _load_ris(self, load_operation: colrev.ops.load.Load) -> dict:
        def entrytype_setter(record_dict: dict) -> None:
            if record_dict["TY"] == "JOUR":
                record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.ARTICLE
            elif record_dict["TY"] == "CONF":
                record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.INPROCEEDINGS
            else:
                record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.MISC

        def field_mapper(record_dict: dict) -> None:

            key_maps = {
                ENTRYTYPES.ARTICLE: {
                    "PY": Fields.YEAR,
                    "AU": Fields.AUTHOR,
                    "TI": Fields.TITLE,
                    "T2": Fields.JOURNAL,
                    "AB": Fields.ABSTRACT,
                    "VL": Fields.VOLUME,
                    "IS": Fields.NUMBER,
                    "DO": Fields.DOI,
                    "PB": Fields.PUBLISHER,
                    "UR": Fields.URL,
                    "SN": Fields.ISSN,
                },
                ENTRYTYPES.INPROCEEDINGS: {
                    "PY": Fields.YEAR,
                    "AU": Fields.AUTHOR,
                    "TI": Fields.TITLE,
                    "T2": Fields.BOOKTITLE,
                    "DO": Fields.DOI,
                    "SN": Fields.ISSN,
                },
            }

            key_map = key_maps[record_dict[Fields.ENTRYTYPE]]
            for ris_key in list(record_dict.keys()):
                if ris_key in key_map:
                    standard_key = key_map[ris_key]
                    record_dict[standard_key] = record_dict.pop(ris_key)

            if "SP" in record_dict and "EP" in record_dict:
                record_dict[Fields.PAGES] = (
                    f"{record_dict.pop('SP')}--{record_dict.pop('EP')}"
                )

            if Fields.AUTHOR in record_dict and isinstance(
                record_dict[Fields.AUTHOR], list
            ):
                record_dict[Fields.AUTHOR] = " and ".join(record_dict[Fields.AUTHOR])
            if Fields.EDITOR in record_dict and isinstance(
                record_dict[Fields.EDITOR], list
            ):
                record_dict[Fields.EDITOR] = " and ".join(record_dict[Fields.EDITOR])
            if Fields.KEYWORDS in record_dict and isinstance(
                record_dict[Fields.KEYWORDS], list
            ):
                record_dict[Fields.KEYWORDS] = ", ".join(record_dict[Fields.KEYWORDS])

            record_dict.pop("TY", None)
            record_dict.pop("Y2", None)
            record_dict.pop("DB", None)
            record_dict.pop("C1", None)
            record_dict.pop("T3", None)
            record_dict.pop("AD", None)
            record_dict.pop("CY", None)
            record_dict.pop("M3", None)
            record_dict.pop("EP", None)
            record_dict.pop("Y1", None)
            record_dict.pop("Y2", None)
            record_dict.pop("JA", None)
            record_dict.pop("JO", None)
            record_dict.pop("VO", None)
            record_dict.pop("VL", None)
            record_dict.pop("IS", None)
            record_dict.pop("ER", None)

            for key, value in record_dict.items():
                record_dict[key] = str(value)

        load_operation.ensure_append_only(self.search_source.filename)
        records = colrev.loader.load_utils.load(
            filename=self.search_source.filename,
            unique_id_field="INCREMENTAL",
            entrytype_setter=entrytype_setter,
            field_mapper=field_mapper,
            logger=self.review_manager.logger,
        )

        return records

    def _load_csv(self) -> dict:
        def entrytype_setter(record_dict: dict) -> None:
            if record_dict["Category"] == "Magazine":
                record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.ARTICLE
            elif record_dict["Category"] == "Conference Publication":
                record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.INPROCEEDINGS
            else:
                record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.MISC

        def field_mapper(record_dict: dict) -> None:
            record_dict[Fields.TITLE] = record_dict.pop("Title", "")
            record_dict[Fields.AUTHOR] = record_dict.pop("Authors", "")
            record_dict[Fields.ABSTRACT] = record_dict.pop("Abstract", "")
            record_dict[Fields.DOI] = record_dict.pop("DOI", "")

            if record_dict[Fields.ENTRYTYPE] == ENTRYTYPES.ARTICLE:
                record_dict[Fields.JOURNAL] = record_dict.pop("Publication", "")
            if record_dict[Fields.ENTRYTYPE] == ENTRYTYPES.INPROCEEDINGS:
                record_dict[Fields.BOOKTITLE] = record_dict.pop("Publication", "")

            record_dict.pop("Category", None)
            record_dict.pop("Affiliations", None)

            for key in list(record_dict.keys()):
                value = record_dict[key]
                record_dict[key] = str(value)
                if value == "" or pd.isna(value):
                    del record_dict[key]

        records = colrev.loader.load_utils.load(
            filename=self.search_source.filename,
            unique_id_field="DOI",
            entrytype_setter=entrytype_setter,
            field_mapper=field_mapper,
            logger=self.review_manager.logger,
        )
        return records

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        if self.search_source.filename.suffix == ".ris":
            return self._load_ris(load_operation)

        if self.search_source.filename.suffix == ".csv":
            return self._load_csv()

        raise NotImplementedError

    def prepare(
        self, record: colrev.record.record.Record, source: colrev.settings.SearchSource
    ) -> colrev.record.record.Record:
        """Source-specific preparation for IEEEXplore"""

        if source.filename.suffix == ".csv":
            if Fields.AUTHOR in record.data:
                record.data[Fields.AUTHOR] = (
                    colrev.record.record_prep.PrepRecord.format_author_field(
                        record.data[Fields.AUTHOR]
                    )
                )
            return record
        return record



#!/usr/bin/env python3
"""Check operation."""
from __future__ import annotations

import typing

import colrev.process.operation
from colrev.constants import OperationsType


if typing.TYPE_CHECKING:  # pragma: no cover
    import colrev.review_manager


class CheckOperation(colrev.process.operation.Operation):
    """A dummy operation that is not expected to introduce changes"""

    # pylint: disable=too-few-public-methods

    type = OperationsType.check

    def __init__(self, review_manager: colrev.review_manager.ReviewManager) -> None:
        super().__init__(
            review_manager=review_manager,
            operations_type=self.type,
            notify_state_transition_operation=False,
        )



#! /usr/bin/env python
"""CoLRev search feed: store and update origin records and update main records."""
from __future__ import annotations

import json
import time
from copy import deepcopy
from random import randint

import colrev.exceptions as colrev_exceptions
import colrev.loader.load_utils
import colrev.loader.load_utils_formatter
import colrev.record.record_merger
from colrev.constants import Colors
from colrev.constants import DefectCodes
from colrev.constants import ENTRYTYPES
from colrev.constants import Fields
from colrev.constants import FieldSet
from colrev.constants import FieldValues
from colrev.writer.write_utils import to_string
from colrev.writer.write_utils import write_file


# Keep in mind the need for lock-mechanisms, e.g., in concurrent prep operations
class SearchAPIFeed:
    """A feed managing results from API searches"""

    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-arguments

    _nr_added: int = 0
    _nr_changed: int = 0

    def __init__(
        self,
        *,
        review_manager: colrev.review_manager.ReviewManager,
        source_identifier: str,
        search_source: colrev.settings.SearchSource,
        update_only: bool,
        prep_mode: bool = False,
    ):
        self.source = search_source
        self.feed_file = search_source.filename

        # Note: the source_identifier identifies records in the search feed.
        # This could be a doi or link or database-specific ID (like WOS accession numbers)
        # The source_identifier can be stored in the main records.bib (it does not have to)
        # The record source_identifier (feed-specific) is used in search
        # or other operations (like prep)
        # In search operations, records are added/updated based on available_ids
        # (which maps source_identifiers to IDs used to generate the colrev_origin)
        # In other operations, records are linked through colrev_origins,
        # i.e., there is no need to store the source_identifier in the main records (redundantly)
        self.source_identifier = source_identifier

        # Note: corresponds to rerun (in search.main() and run_search())
        self.update_only = update_only
        # if update_only, we do not update time_variant_fields
        # (otherwise, fields in recent records would be more up-to-date)

        self.review_manager = review_manager
        self.logger = review_manager.logger
        self.load_formatter = colrev.loader.load_utils_formatter.LoadFormatter()

        self.origin_prefix = self.source.get_origin_prefix()

        self._load_feed()

        self.prep_mode = prep_mode
        if not prep_mode:
            self.records = self.review_manager.dataset.load_records_dict()

    def get_last_updated(self) -> str:
        """Returns the date of the last update (if available) in YYYY-MM-DD format"""
        file = self.feed_file
        if not file.is_file():
            return ""
        return self.review_manager.dataset.get_last_commit_date(self.feed_file)

    def _load_feed(self) -> None:
        if not self.feed_file.is_file():
            self._available_ids = {}
            self._next_incremental_id = 1
            self.feed_records = {}
            return
        self.feed_records = colrev.loader.load_utils.loads(
            load_string=self.feed_file.read_text(encoding="utf8"),
            implementation="bib",
            logger=self.review_manager.logger,
        )
        self._available_ids = {
            x[self.source_identifier]: x[Fields.ID]
            for x in self.feed_records.values()
            if self.source_identifier in x
        }
        self._next_incremental_id = (
            max(
                [
                    int(x[Fields.ID])
                    for x in self.feed_records.values()
                    if x[Fields.ID].isdigit()
                ]
                + [1]
            )
            + 1
        )

    def _set_id(self, record: colrev.record.record.Record) -> None:
        """Set incremental record ID
        If self.source_identifier is in record_dict, it is updated, otherwise added as a new record.
        """

        if self.source_identifier not in record.data:
            raise colrev_exceptions.NotFeedIdentifiableException()

        if record.data[self.source_identifier] in self._available_ids:
            record.data[Fields.ID] = self._available_ids[
                record.data[self.source_identifier]
            ]
        else:
            record.data[Fields.ID] = str(self._next_incremental_id).rjust(6, "0")

    def add_record_to_feed(
        self,
        record: colrev.record.record.Record,
        prev_feed_record: colrev.record.record.Record,
    ) -> bool:
        """Add a record to the feed and set its colrev_origin"""

        self._set_id(record)

        feed_record_dict = record.data.copy()
        added_new = True
        if feed_record_dict[self.source_identifier] in self._available_ids:
            added_new = False
        else:
            self._next_incremental_id += 1
            self._nr_added += 1

        frid = feed_record_dict[Fields.ID]
        self._available_ids[feed_record_dict[self.source_identifier]] = frid

        self._notify_record_forthcoming(
            record=record, prev_feed_record=prev_feed_record
        )

        if self.update_only:
            for key in FieldSet.TIME_VARIANT_FIELDS:
                if frid in self.feed_records:
                    feed_record_dict.pop(key, None)
                    if key in self.feed_records[frid]:
                        feed_record_dict[key] = self.feed_records[frid][key]

        self.feed_records[frid] = feed_record_dict
        if added_new:
            if not self.prep_mode:
                self.logger.info(f"  add record: {record.data[self.source_identifier]}")
            elif (
                self.prep_mode and self.review_manager.verbose_mode
            ):  # pragma: no cover
                self.logger.info(
                    f"  link record: {record.data[self.source_identifier]}"
                )
        return added_new

    def _have_changed(
        self,
        record_a: colrev.record.record.Record,
        record_b: colrev.record.record.Record,
    ) -> bool:

        def tostr_and_load(record: colrev.record.record.Record) -> dict:
            record_dict = deepcopy(record.data)
            bibtex_str = to_string(
                records_dict={record_dict[Fields.ID]: record_dict}, implementation="bib"
            )
            record_dict = list(
                colrev.loader.load_utils.loads(
                    load_string=bibtex_str,
                    implementation="bib",
                    logger=self.review_manager.logger,
                ).values()
            )[0]
            return record_dict

        # To ignore changes introduced by saving/loading the feed-records,
        # we parse and load them in the following.
        record_a_dict = tostr_and_load(record_a)
        record_b_dict = tostr_and_load(record_b)

        # Note : record_a can have more keys (that's ok)
        changed = False
        for key, value in record_b_dict.items():
            if key in FieldSet.PROVENANCE_KEYS + [Fields.ID, Fields.CURATION_ID]:
                continue
            if key not in record_a_dict or record_a_dict[key] != value:
                return True
        return changed

    def _update_record_retract(
        self,
        *,
        record: colrev.record.record.Record,
        main_record: colrev.record.record.Record,
    ) -> bool:
        if record.is_retracted():
            self.review_manager.logger.info(
                f"{Colors.RED}Found paper retract: "
                f"{main_record.data['ID']}{Colors.END}"
            )
            main_record.prescreen_exclude(
                reason=FieldValues.RETRACTED, print_warning=True
            )
            main_record.remove_field(key="warning")
            return True
        return False

    def _notify_record_forthcoming(
        self,
        *,
        record: colrev.record.record.Record,
        prev_feed_record: colrev.record.record.Record,
    ) -> None:

        if self._forthcoming_published(record=record, prev_record=prev_feed_record):
            self.review_manager.logger.info(
                f"{Colors.GREEN}Update published forthcoming paper: "
                f"{record.data['ID']}{Colors.END}"
            )

    def _missing_ignored_field(
        self, main_record: colrev.record.record.Record, key: str
    ) -> bool:
        source = main_record.get_field_provenance_source(key)
        notes = main_record.get_field_provenance_notes(key)
        if (
            source == "colrev_curation.masterdata_restrictions"
            and f"IGNORE:{DefectCodes.MISSING}" in notes
        ):
            return True
        return False

    def _update_record_fields(
        self,
        *,
        record: colrev.record.record.Record,
        main_record: colrev.record.record.Record,
    ) -> None:

        self._update_record_retract(record=record, main_record=main_record)

        if main_record.masterdata_is_curated() and not self.source.is_curated_source():
            return

        for key in list(record.data.keys()):
            if self.update_only and key in FieldSet.TIME_VARIANT_FIELDS:
                continue

            if key in FieldSet.PROVENANCE_KEYS + [Fields.ID, Fields.CURATION_ID]:
                continue

            if key not in main_record.data:
                if self._missing_ignored_field(main_record, key):
                    continue
                colrev.record.record_merger.fuse_fields(
                    main_record, merging_record=record, key=key
                )

            else:
                # Ignore minor changes
                if record.data[key].replace(" - ", ": ") == main_record.data[
                    key
                ].replace(" - ", ": "):
                    continue
                colrev.record.record_merger.fuse_fields(
                    main_record, merging_record=record, key=key
                )

    def _forthcoming_published(
        self,
        *,
        record: colrev.record.record.Record,
        prev_record: colrev.record.record.Record,
    ) -> bool:
        if (
            record.data[Fields.ENTRYTYPE] != ENTRYTYPES.ARTICLE
            or Fields.YEAR not in record.data
            or Fields.YEAR not in prev_record.data
        ):
            return False

        # Option 1: Forthcoming paper published if year is assigned
        if (
            "forthcoming" == prev_record.data[Fields.YEAR]
            and "forthcoming" != record.data[Fields.YEAR]
        ):
            return True

        # Option 2: Forthcoming paper published if volume and number are assigned
        # i.e., no longer UNKNOWN
        if (
            record.data.get(Fields.VOLUME, FieldValues.UNKNOWN) != FieldValues.UNKNOWN
            and prev_record.data.get(Fields.VOLUME, FieldValues.UNKNOWN)
            == FieldValues.UNKNOWN
            and record.data.get(Fields.NUMBER, FieldValues.UNKNOWN)
            != FieldValues.UNKNOWN
            and prev_record.data.get(Fields.VOLUME, FieldValues.UNKNOWN)
            == FieldValues.UNKNOWN
        ):
            return True
        return False

    def _get_main_record(self, colrev_origin: str) -> colrev.record.record.Record:

        main_record_dict: dict = {}
        for record_dict in self.records.values():
            if colrev_origin in record_dict[Fields.ORIGIN]:
                main_record_dict = record_dict
                break

        if main_record_dict == {}:
            raise colrev_exceptions.RecordNotFoundException(
                f"Could not find/update {colrev_origin}"
            )
        return colrev.record.record.Record(main_record_dict)

    def _update_record(
        self,
        *,
        retrieved_record: colrev.record.record.Record,
        prev_feed_record: colrev.record.record.Record,
    ) -> bool:
        """Convenience function to update existing records (main data/records.bib)"""

        colrev_origin = f"{self.origin_prefix}/{retrieved_record.data['ID']}"
        main_record = self._get_main_record(colrev_origin)

        # For consistency (with ops/load):
        self.load_formatter.run(retrieved_record)

        self._update_record_fields(
            record=retrieved_record,
            main_record=main_record,
        )

        if self._forthcoming_published(
            record=retrieved_record, prev_record=prev_feed_record
        ):
            # (notified when updating feed record)
            return True

        if self._have_changed(retrieved_record, prev_feed_record):
            similarity_score = colrev.record.record.Record.get_record_similarity(
                retrieved_record,
                prev_feed_record,
            )
            self._nr_changed += 1
            if similarity_score > 0.98:
                self.review_manager.logger.info(f" check/update {colrev_origin}")
            else:
                dict_diff = retrieved_record.get_diff(prev_feed_record)
                self.review_manager.logger.info(
                    f" {Colors.RED} check/update {colrev_origin} leads to substantial changes "
                    f"({similarity_score}) in {main_record.data['ID']}:{Colors.END}"
                )
                self.review_manager.logger.info(
                    self.review_manager.p_printer.pformat(
                        [x for x in dict_diff if "change" == x[0]]
                    )
                )
            return True
        return False

    def _print_post_run_search_infos(self) -> None:
        """Print the search infos (after running the search)"""
        if self._nr_added > 0:
            self.review_manager.logger.info(
                f"{Colors.GREEN}Retrieved {self._nr_added} records{Colors.END}"
            )
        else:
            self.review_manager.logger.info(
                f"{Colors.GREEN}No additional records retrieved{Colors.END}"
            )

        if self.prep_mode:  # pragma: no cover
            # No need to print the following in prep mode
            # because the main records are not updated/saved
            return

        if self._nr_changed > 0:  # pragma: no cover
            self.review_manager.logger.info(
                f"{Colors.GREEN}Updated {self._nr_changed} records{Colors.END}"
            )
        else:
            if self.records:
                self.review_manager.logger.info(
                    f"{Colors.GREEN}Records ({self.review_manager.paths.RECORDS_FILE})"
                    f" up-to-date{Colors.END}"
                )

    def get_prev_feed_record(
        self, record: colrev.record.record.Record
    ) -> colrev.record.record.Record:
        """Get the previous record dict version"""
        record = deepcopy(record)
        self._set_id(record)
        prev_feed_record_dict = {}
        if record.data[Fields.ID] in self.feed_records:
            prev_feed_record_dict = deepcopy(self.feed_records[record.data[Fields.ID]])
        return colrev.record.record.Record(prev_feed_record_dict)

    def _prep_retrieved_record(
        self, retrieved_record: colrev.record.record.Record
    ) -> None:
        """Prepare the retrieved record for the search feed"""
        for provenance_key in FieldSet.PROVENANCE_KEYS:
            if provenance_key in retrieved_record.data:
                del retrieved_record.data[provenance_key]

    def add_update_record(self, retrieved_record: colrev.record.record.Record) -> bool:
        """Add or update a record in the api_search_feed and records"""
        self._prep_retrieved_record(retrieved_record)
        prev_feed_record = self.get_prev_feed_record(retrieved_record)

        added = self.add_record_to_feed(retrieved_record, prev_feed_record)
        updated = False
        if not self.prep_mode:
            try:
                updated = self._update_record(
                    retrieved_record=retrieved_record.copy(),
                    prev_feed_record=prev_feed_record.copy(),
                )
            except colrev_exceptions.RecordNotFoundException:
                pass

        if self.prep_mode:
            retrieved_record.data[Fields.ORIGIN] = [
                f"{self.origin_prefix}/{retrieved_record.data['ID']}"
            ]
        return added or updated

    def save(self, *, skip_print: bool = False) -> None:
        """Save the feed file and records, printing post-run search infos."""

        if not skip_print and not self.prep_mode:
            self._print_post_run_search_infos()

        if len(self.feed_records) > 0:
            self.feed_file.parents[0].mkdir(parents=True, exist_ok=True)
            write_file(records_dict=self.feed_records, filename=self.feed_file)

            while True:
                try:
                    self.review_manager.load_settings()
                    if self.source.filename.name not in [
                        s.filename.name for s in self.review_manager.settings.sources
                    ]:
                        self.review_manager.settings.sources.append(self.source)
                        self.review_manager.save_settings()

                    self.review_manager.dataset.add_changes(self.feed_file)
                    break
                except (
                    FileExistsError,
                    OSError,
                    json.decoder.JSONDecodeError,
                ):  # pragma: no cover
                    self.review_manager.logger.debug("Wait for git")
                    time.sleep(randint(1, 15))  # nosec

        if not self.prep_mode:
            self.review_manager.dataset.save_records_dict(self.records)
        if not skip_print:
            self._nr_added = 0
            self._nr_changed = 0



#! /usr/bin/env python
"""Checker for fields that are inconsistent with the ENTRYTYPE."""
from __future__ import annotations

import colrev.record.qm.quality_model
from colrev.constants import DefectCodes
from colrev.constants import Fields

# pylint: disable=too-few-public-methods


class InconsistentWithEntrytypeChecker:
    """The InconsistentWithEntrytypeChecker"""

    record_field_inconsistencies: dict[str, list[str]] = {
        "article": [Fields.BOOKTITLE, Fields.ISBN],
        "inproceedings": ["issue", Fields.NUMBER, Fields.JOURNAL],
        "incollection": [],
        "inbook": [Fields.JOURNAL],
        "book": ["issue", Fields.NUMBER, Fields.JOURNAL],
        "phdthesis": [
            Fields.VOLUME,
            "issue",
            Fields.NUMBER,
            Fields.JOURNAL,
            Fields.BOOKTITLE,
            Fields.ISBN,
        ],
        "masterthesis": [
            Fields.VOLUME,
            "issue",
            Fields.NUMBER,
            Fields.JOURNAL,
            Fields.BOOKTITLE,
            Fields.ISBN,
        ],
        "techreport": [
            Fields.VOLUME,
            "issue",
            Fields.NUMBER,
            Fields.JOURNAL,
            Fields.BOOKTITLE,
            Fields.ISBN,
        ],
        "unpublished": [
            Fields.VOLUME,
            "issue",
            Fields.NUMBER,
            Fields.JOURNAL,
            Fields.BOOKTITLE,
            Fields.ISBN,
        ],
        "online": [Fields.JOURNAL, Fields.BOOKTITLE, Fields.ISBN],
        "misc": [Fields.JOURNAL, Fields.BOOKTITLE, Fields.ISBN],
    }
    """Fields considered inconsistent with the respective ENTRYTYPE"""

    msg = DefectCodes.INCONSISTENT_WITH_ENTRYTYPE

    def __init__(
        self, quality_model: colrev.record.qm.quality_model.QualityModel
    ) -> None:
        self.quality_model = quality_model

    def run(self, *, record: colrev.record.record.Record) -> None:
        """Run the completeness checks"""

        if "ENTRYTYPE" not in record.data:
            return

        if record.data["ENTRYTYPE"] not in self.record_field_inconsistencies:
            return

        inconsistent_fields = self.record_field_inconsistencies[
            record.data["ENTRYTYPE"]
        ]
        for key in list(record.data.keys()):
            if record.ignored_defect(key=key, defect=self.msg):
                continue
            if key in inconsistent_fields:
                record.add_field_provenance_note(key=key, note=self.msg)
            else:
                record.remove_field_provenance_note(key=key, note=self.msg)


def register(quality_model: colrev.record.qm.quality_model.QualityModel) -> None:
    """Register the checker"""
    quality_model.register_checker(InconsistentWithEntrytypeChecker(quality_model))



#! /usr/bin/env python
"""DBLP API"""
import html
import json
import re
import time
from datetime import datetime

import requests

import colrev.exceptions as colrev_exceptions
import colrev.record.record_prep
from colrev.constants import ENTRYTYPES
from colrev.constants import Fields

# pylint: disable=too-few-public-methods
# pylint: disable=too-many-instance-attributes


class DBLPAPI:
    """Connector for the DBLP API"""

    # https://dblp.org/search/publ/api?q=ADD_TITLE&format=json
    _api_url_venues = "https://dblp.org/search/venue/api?q="
    _api_url = "https://dblp.org/search/publ/api?q="

    url = ""
    _batch_next = False

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        *,
        params: dict,
        email: str,
        session: requests.Session,
        rerun: bool = False,
        timeout: int = 60,
    ):
        self.params = params
        self.email = email
        self.session = session
        self._timeout = timeout

        self.headers = {"user-agent": f"{__name__}  (mailto:{self.email})"}

        self.batch_size_cumulative = 0
        self.batch_size = 250
        self.year = datetime.now().year - 2
        if rerun:
            self.year = 1980
        self.total = self.get_total()

    def check_availability(self) -> None:
        """Check if the DBLP API is available"""

        try:
            # pylint: disable=duplicate-code
            test_rec = {
                Fields.ENTRYTYPE: "article",
                Fields.DOI: "10.17705/1cais.04607",
                Fields.AUTHOR: "Schryen, Guido and Wagner, Gerit and Benlian, Alexander "
                "and Paré, Guy",
                Fields.TITLE: "A Knowledge Development Perspective on Literature Reviews: "
                "Validation of a new Typology in the IS Field",
                Fields.ID: "SchryenEtAl2021",
                Fields.JOURNAL: "Communications of the Association for Information Systems",
                Fields.VOLUME: "46",
                Fields.YEAR: "2020",
                # Fields.STATUS: RecordState.md_prepared,  # type: ignore
            }
            self.params = {"query": str(test_rec[Fields.TITLE])}
            self.set_url_from_query()

            retrieved_records = self.retrieve_records()
            dblp_record = retrieved_records[0]

            if 0 != len(dblp_record.data):
                assert dblp_record.data[Fields.TITLE] == test_rec[Fields.TITLE]
                assert dblp_record.data[Fields.AUTHOR] == test_rec[Fields.AUTHOR]
            else:
                raise colrev_exceptions.ServiceNotAvailableException("DBLP")
        except requests.exceptions.RequestException as exc:
            raise colrev_exceptions.ServiceNotAvailableException("DBLP") from exc

    def _get_dblp_venue(
        self,
        venue_string: str,
        *,
        venue_type: str,
    ) -> str:
        # Note : venue_string should be like "behaviourIT"
        # Note : journals that have been renamed seem to return the latest
        # journal name. Example:
        # https://dblp.org/db/journals/jasis/index.html
        venue = venue_string
        url = self._api_url_venues + venue_string.replace(" ", "+") + "&format=json"
        try:
            ret = self.session.request(
                "GET", url, headers=self.headers, timeout=self._timeout
            )
            ret.raise_for_status()
            data = json.loads(ret.text)
            if "hit" not in data["result"]["hits"]:
                return ""
            hits = data["result"]["hits"]["hit"]
            for hit in hits:
                if hit["info"]["type"] != venue_type:
                    continue
                # pylint: disable=colrev-missed-constant-usage
                if f"/{venue_string.lower()}/" in hit["info"]["url"].lower():
                    venue = hit["info"]["venue"]
                    break

            venue = re.sub(r" \(.*?\)", "", venue)
        except requests.exceptions.RequestException:
            pass
        return venue

    # pylint: disable=too-many-branches
    def _dblp_json_set_type(self, *, item: dict) -> None:
        lpos = item["key"].find("/") + 1
        rpos = item["key"].rfind("/")
        ven_key = item["key"][lpos:rpos]

        if item["type"] == "Withdrawn Items":
            if item["key"][:8] == "journals":
                item["type"] = "Journal Articles"
            if item["key"][:4] == "conf":
                item["type"] = "Conference and Workshop Papers"
            item["colrev.dblp.warning"] = "Withdrawn (according to DBLP)"

        if "corr" == ven_key:
            item[Fields.ENTRYTYPE] = ENTRYTYPES.TECHREPORT

        elif item["type"] == "Journal Articles":
            item[Fields.ENTRYTYPE] = ENTRYTYPES.ARTICLE
            item[Fields.JOURNAL] = self._get_dblp_venue(
                ven_key,
                venue_type="Journal",
            )
        elif item["type"] == "Conference and Workshop Papers":
            item[Fields.ENTRYTYPE] = ENTRYTYPES.INPROCEEDINGS
            item[Fields.BOOKTITLE] = self._get_dblp_venue(
                ven_key,
                venue_type="Conference or Workshop",
            )
        elif item["type"] == "Informal and Other Publications":
            item[Fields.ENTRYTYPE] = ENTRYTYPES.MISC
            item[Fields.BOOKTITLE] = item["venue"]
        elif item["type"] == "Parts in Books or Collections":
            item[Fields.ENTRYTYPE] = ENTRYTYPES.INBOOK
            item[Fields.BOOKTITLE] = item["venue"]
        elif item["type"] == "Books and Theses":
            if item["key"].startswith("phd/"):
                item[Fields.ENTRYTYPE] = ENTRYTYPES.PHDTHESIS
            else:
                item[Fields.ENTRYTYPE] = ENTRYTYPES.BOOK

        else:
            item[Fields.ENTRYTYPE] = ENTRYTYPES.MISC
            if item["type"] != "Editorship":
                print("DBLP: Unknown type: %s", item)

    def _dblp_json_to_dict(
        self,
        *,
        item: dict,
    ) -> dict:
        # pylint: disable=too-many-branches
        # To test in browser:
        # https://dblp.org/search/publ/api?q=ADD_TITLE&format=json

        self._dblp_json_set_type(item=item)
        if "title" in item:
            item[Fields.TITLE] = item["title"].rstrip(".").rstrip().replace("\n", " ")
            item[Fields.TITLE] = re.sub(r"\s+", " ", item[Fields.TITLE])
        # Note : DBLP provides number-of-pages (instead of pages start-end)
        # if "pages" in item:
        #     item[Fields.PAGES] = item[Fields.PAGES].replace("-", "--")
        if "authors" in item:
            if Fields.AUTHOR in item["authors"]:
                if isinstance(item["authors"][Fields.AUTHOR], dict):
                    author_string = item["authors"][Fields.AUTHOR]["text"]
                else:
                    authors_nodes = [
                        author
                        for author in item["authors"][Fields.AUTHOR]
                        if isinstance(author, dict)
                    ]
                    authors = [x["text"] for x in authors_nodes if "text" in x]
                    author_string = " and ".join(authors)
                author_string = (
                    colrev.record.record_prep.PrepRecord.format_author_field(
                        author_string
                    )
                )
                item[Fields.AUTHOR] = author_string

        if "key" in item:
            item["dblp_key"] = "https://dblp.org/rec/" + item["key"]

        if Fields.DOI in item:
            item[Fields.DOI] = item[Fields.DOI].upper()
        if "ee" in item:
            if not any(
                x in item["ee"] for x in ["https://doi.org", "https://dblp.org"]
            ):
                item[Fields.URL] = item["ee"]
        if Fields.URL in item:
            if "https://dblp.org" in item[Fields.URL]:
                del item[Fields.URL]

        item = {
            k: v
            for k, v in item.items()
            if k not in ["venue", "type", "access", "key", "ee", "authors"]
        }
        for key, value in item.items():
            item[key] = html.unescape(value).replace("{", "").replace("}", "")

        return item

    def get_query_url(self) -> str:
        """Get the query"""

        if "scope" in self.params:
            # Note : journal_abbreviated is the abbreviated venue_key
            query = self.params["scope"]["journal_abbreviated"]
            # query = params['scope']["venue_key"]
        elif "query" in self.params:
            query = self.params["query"]
        else:
            raise ValueError("No query or scope provided")

        return self._api_url + query.replace(" ", "+")

    def set_next_url(self) -> None:
        """Set the next URL"""

        self.url = self.get_query_url()
        if self.total > self.batch_size:
            self.url += "+" + str(self.year)
            if self._batch_next:
                self.batch_size_cumulative += self.batch_size
            else:
                self.batch_size_cumulative = 0
            self.year += 1

        self.url += f"&format=json&h={self.batch_size}&f={self.batch_size_cumulative}"

    def processed_all_urls(self) -> bool:
        """Check if all URLs have been processed"""

        if self.total < self.batch_size:
            return True
        return self.year > datetime.now().year

    def get_total(self) -> int:
        """Get the total number of records"""

        try:
            ret = self.session.request(
                "GET",
                self.get_query_url() + "&format=json",
                headers=self.headers,
                timeout=self._timeout,
            )
            ret.raise_for_status()

            data = json.loads(ret.text)

            if "result" not in data:
                return -1
            if "hits" not in data["result"]:
                return -1
            if "@total" not in data["result"]["hits"]:
                return -1
            return int(data["result"]["hits"]["@total"])
        except (requests.exceptions.RequestException, ValueError):
            return -1

    def retrieve_records(self) -> list:
        """Retrieve records from DBLP"""

        # try:
        while True:
            # review_manager.logger.debug(url)
            ret = self.session.request(
                "GET", self.url, headers=self.headers, timeout=self._timeout  # type: ignore
            )

            if ret.status_code == 429:
                time.sleep(60)
                print("Waiting for 60 seconds (request limit reached)")
                continue
            ret.raise_for_status()
            # 429 - too many requests
            if ret.status_code == 500:
                return []
            break

        data = json.loads(ret.text)
        response_ms = float(data["result"]["time"]["text"])
        time.sleep(response_ms / 10)

        if "hits" not in data["result"]:
            return []
        if "hit" not in data["result"]["hits"]:
            return []
        hits = data["result"]["hits"]["hit"]
        items = [hit["info"] for hit in hits]

        dblp_dicts = [
            self._dblp_json_to_dict(
                item=item,
            )
            for item in items
        ]
        if len(dblp_dicts) > self.batch_size:
            self._batch_next = True
        else:
            self._batch_next = False

        retrieved_records = [
            colrev.record.record_prep.PrepRecord(dblp_dict) for dblp_dict in dblp_dicts
        ]

        for retrieved_record in retrieved_records:
            retrieved_record.add_provenance_all(
                source=retrieved_record.data["dblp_key"]
            )

        # # pylint: disable=duplicate-code
        # except OperationalError as exc:
        #     raise colrev_exceptions.ServiceNotAvailableException(
        #         "sqlite, required for requests CachedSession "
        #         "(possibly caused by concurrent operations)"
        #     ) from exc
        # except (requests.exceptions.ReadTimeout, requests.exceptions.HTTPError) as exc:
        #     raise colrev_exceptions.ServiceNotAvailableException(
        #         "requests timed out "
        #         "(possibly because the DBLP service is temporarily not available)"
        #     ) from exc

        return retrieved_records

    def set_url_from_query(self) -> None:
        """Set the URL from a query"""
        query = re.sub(
            r"[\W]+", " ", self.params["query"].replace(" ", "_").replace("-", "_")
        )
        self.url = self._api_url + query.replace(" ", "+") + "&format=json"



#! /usr/bin/env python
"""SearchSource: Transport Research International Documentation"""
from __future__ import annotations

import re
from pathlib import Path

import zope.interface
from pydantic import Field

import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import ENTRYTYPES
from colrev.constants import Fields
from colrev.constants import SearchSourceHeuristicStatus
from colrev.constants import SearchType

# pylint: disable=unused-argument
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.package_manager.interfaces.SearchSourceInterface)
class TransportResearchInternationalDocumentation:
    """Transport Research International Documentation"""

    settings_class = colrev.package_manager.package_settings.DefaultSourceSettings

    endpoint = "colrev.trid"
    source_identifier = "biburl"
    search_types = [SearchType.DB]

    ci_supported: bool = Field(default=False)
    heuristic_status = SearchSourceHeuristicStatus.supported

    db_url = "https://trid.trb.org/"

    def __init__(
        self, *, source_operation: colrev.process.operation.Operation, settings: dict
    ) -> None:
        self.search_source = self.settings_class(**settings)
        self.source_operation = source_operation
        self.review_manager = source_operation.review_manager

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for Transport Research International Documentation"""

        result = {"confidence": 0.0}
        # Simple heuristic:
        if "UR  - https://trid.trb.org/view/" in data:
            result["confidence"] = 0.9
            return result
        return result

    @classmethod
    def add_endpoint(
        cls,
        operation: colrev.ops.search.Search,
        params: str,
    ) -> colrev.settings.SearchSource:
        """Add SearchSource as an endpoint (based on query provided to colrev search --add )"""

        search_source = operation.create_db_source(
            search_source_cls=cls,
            params={},
        )
        operation.add_source_and_search(search_source)
        return search_source

    def search(self, rerun: bool) -> None:
        """Run a search of TRID"""

        if self.search_source.search_type == SearchType.DB:
            self.source_operation.run_db_search(  # type: ignore
                search_source_cls=self.__class__,
                source=self.search_source,
            )
            return

        raise NotImplementedError

    def prep_link_md(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.record.Record:
        """Not implemented"""
        return record

    def _load_ris(self) -> dict:
        def entrytype_setter(record_dict: dict) -> None:
            if record_dict["TY"] in ["JOUR", "ABST"]:
                record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.ARTICLE
            elif record_dict["TY"] == "RPRT":
                record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.TECHREPORT
            elif record_dict["TY"] == "CHAP":
                record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.INBOOK
            else:
                record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.MISC

        def field_mapper(record_dict: dict) -> None:

            key_maps = {
                ENTRYTYPES.ARTICLE: {
                    "PY": Fields.YEAR,
                    "AU": Fields.AUTHOR,
                    "TI": Fields.TITLE,
                    "JO": Fields.JOURNAL,
                    "AB": Fields.ABSTRACT,
                    "VL": Fields.VOLUME,
                    "IS": Fields.NUMBER,
                    "DO": Fields.DOI,
                    "PB": Fields.PUBLISHER,
                    "KW": Fields.KEYWORDS,
                    "UR": Fields.URL,
                    "SP": Fields.PAGES,
                    "AN": f"{self.endpoint}.accession_number",
                },
                ENTRYTYPES.TECHREPORT: {
                    "PY": Fields.YEAR,
                    "AU": Fields.AUTHOR,
                    "TI": Fields.TITLE,
                    "UR": Fields.URL,
                    "PB": Fields.PUBLISHER,
                    "KW": Fields.KEYWORDS,
                    "SP": Fields.PAGES,
                    "AB": Fields.ABSTRACT,
                    "AN": f"{self.endpoint}.accession_number",
                },
            }

            key_map = key_maps[record_dict[Fields.ENTRYTYPE]]
            for ris_key in list(record_dict.keys()):
                if ris_key in key_map:
                    standard_key = key_map[ris_key]
                    record_dict[standard_key] = record_dict.pop(ris_key)

            if "SP" in record_dict and "EP" in record_dict:
                record_dict[Fields.PAGES] = (
                    f"{record_dict.pop('SP')}--{record_dict.pop('EP')}"
                )

            urls = record_dict.get(Fields.URL, [])
            fulltext = [url for url in urls if url.endswith(".pdf")]
            if fulltext:
                urls.remove(fulltext[0])
                record_dict[Fields.FULLTEXT] = fulltext[0]
            trid_url = [url for url in urls if url.startswith("https://trid.trb.org")]
            if trid_url:
                urls.remove(trid_url[0])
                record_dict[f"{self.endpoint}.trid_url"] = trid_url[0]
            record_dict[Fields.URL] = urls[0]

            if Fields.AUTHOR in record_dict and isinstance(
                record_dict[Fields.AUTHOR], list
            ):
                record_dict[Fields.AUTHOR] = " and ".join(record_dict[Fields.AUTHOR])
            if Fields.EDITOR in record_dict and isinstance(
                record_dict[Fields.EDITOR], list
            ):
                record_dict[Fields.EDITOR] = " and ".join(record_dict[Fields.EDITOR])
            if Fields.KEYWORDS in record_dict and isinstance(
                record_dict[Fields.KEYWORDS], list
            ):
                record_dict[Fields.KEYWORDS] = ", ".join(record_dict[Fields.KEYWORDS])

            record_dict.pop("TY", None)
            record_dict.pop("Y2", None)
            record_dict.pop("DB", None)
            record_dict.pop("C1", None)
            record_dict.pop("T3", None)
            record_dict.pop("AD", None)
            record_dict.pop("CY", None)
            record_dict.pop("M3", None)
            record_dict.pop("EP", None)
            record_dict.pop("ET", None)
            record_dict.pop("ER", None)

            for key, value in record_dict.items():
                record_dict[key] = str(value)

        records = colrev.loader.load_utils.load(
            filename=self.search_source.filename,
            unique_id_field="AN",
            entrytype_setter=entrytype_setter,
            field_mapper=field_mapper,
            logger=self.review_manager.logger,
        )

        return records

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        if self.search_source.filename.suffix == ".ris":
            return self._load_ris()

        raise NotImplementedError

    def prepare(
        self, record: colrev.record.record.Record, source: colrev.settings.SearchSource
    ) -> colrev.record.record.Record:
        """Source-specific preparation for Transport Research International Documentation"""

        if Fields.YEAR in record.data:
            if not re.match(r"^\d{4}$", record.data[Fields.YEAR]):
                result = re.search(r"\d{4}", record.data[Fields.YEAR])
                if result:
                    year = result.group(0)
                    if 1900 < int(year) < 2030:
                        record.data[Fields.YEAR] = year

        return record



#!/usr/bin/env python
"""Test the crossref SearchSource"""
from pathlib import Path

import pytest
import requests_mock

import colrev.ops.prep
import colrev.packages.crossref.src.crossref_search_source
from colrev.packages.crossref.src import crossref_api

# pylint: disable=line-too-long


@pytest.mark.parametrize(
    "doi, expected_dict",
    [
        (
            "10.2196/22081",
            {
                "doi": "10.2196/22081",
                "ENTRYTYPE": "article",
                "author": "Grenier Ouimet, Antoine and Wagner, Gerit and Raymond, Louis and "
                "Pare, Guy",
                "journal": "Journal of Medical Internet Research",
                "language": "en",
                "number": "11",
                "pages": "e22081",
                "title": "Investigating Patients’ Intention to Continue Using "
                "Teleconsultation to Anticipate Postcrisis Momentum: Survey Study",
                "abstract": "Background The COVID-19 crisis has drastically changed care "
                "delivery with teleconsultation platforms experiencing "
                "substantial spikes in demand, helping patients and care "
                "providers avoid infections and maintain health care services. "
                "Beyond the current pandemic, teleconsultation is considered a "
                "significant opportunity to address persistent health system "
                "challenges, including accessibility, continuity, and cost of "
                "care, while ensuring quality. Objective This study aims at "
                "identifying the determinants of patients’ intention to continue "
                "using a teleconsultation platform. It extends prior research on "
                "information technology use continuance intention and "
                "teleconsultation services. Methods Data was collected in "
                "November 2018 and May 2019 with Canadian patients who had access "
                "to a teleconsultation platform. Measures included patients’ "
                "intention to continue their use; teleconsultation usefulness; "
                "teleconsultation quality; patients’ trust toward the digital "
                "platform, its provider. and health care professionals; and "
                "confirmation of patients’ expectations toward teleconsultation. "
                "We used structural equation modeling employing the partial least "
                "squares component-based technique to test our research model and "
                "hypotheses. Results We analyzed a sample of 178 participants who "
                "had used teleconsultation services. Our findings revealed that "
                "confirmation of expectations had the greatest influence on "
                "continuance intention (total effects=0.722; P<.001), followed by "
                "usefulness (total effects=0.587; P<.001) and quality (total "
                "effects=0.511; P<.001). Usefulness (β=.60; P<.001) and quality "
                "(β=.34; P=.01) had direct effects on the dependent variable. The "
                "confirmation of expectations had direct effects both on "
                "usefulness (β=.56; P<.001) and quality (β=.75; P<.001) in "
                "addition to having an indirect effect on usefulness (indirect "
                "effects=0.282; P<.001). Last, quality directly influenced "
                "usefulness (β=.34; P=.002) and trust (β=.88; P<.001). Trust does "
                "not play a role in the context under study. Conclusions "
                "Teleconsultation is central to care going forward, and it "
                "represents a significant lever for an improved, digital delivery "
                "of health care in the future. We believe that our findings will "
                "help drive long-term teleconsultation adoption and use, "
                "including in the aftermath of the current COVID-19 crisis, so "
                "that general care improvement and greater preparedness for "
                "exceptional situations can be achieved.",
                "volume": "22",
                "year": "2020",
            },
        ),
        (
            "10.17705/1cais.04607",
            {
                "doi": "10.17705/1CAIS.04607",
                "ENTRYTYPE": "article",
                "author": "Schryen, Guido and Wagner, Gerit and Benlian, Alexander and Paré, Guy",
                "title": "A Knowledge Development Perspective on Literature Reviews: Validation of a new Typology in the IS Field",  # noqa: E501
                "journal": "Communications of the Association for Information Systems",
                "volume": "49",
                "year": "2021",
                "number": "1",
                "pages": "134--186",
            },
        ),
        (
            "10.1177/02683962211048201",
            {
                "doi": "10.1177/02683962211048201",
                "ENTRYTYPE": "article",
                "abstract": "Artificial intelligence (AI) is beginning to transform traditional "
                + "research practices in many areas. In this context, literature "
                + "reviews stand out because they operate on large and rapidly "
                + "growing volumes of documents, that is, partially structured "
                + "(meta)data, and pervade almost every type of paper published in "
                + "information systems research or related social science "
                + "disciplines. To familiarize researchers with some of the recent "
                + "trends in this area, we outline how AI can expedite individual "
                + "steps of the literature review process. Considering that the use "
                + "of AI in this context is in an early stage of development, we "
                + "propose a comprehensive research agenda for AI-based literature "
                + "reviews (AILRs) in our field. With this agenda, we would like to "
                + "encourage design science research and a broader constructive "
                + "discourse on shaping the future of AILRs in research.",
                "author": "Wagner, Gerit and Lukyanenko, Roman and Paré, Guy",
                "fulltext": "http://journals.sagepub.com/doi/pdf/10.1177/02683962211048201",
                "journal": "Journal of Information Technology",
                "language": "en",
                "number": "2",
                "pages": "209--226",
                "title": "Artificial intelligence and the conduct of literature reviews",
                "volume": "37",
                "year": "2022",
            },
        ),
    ],
)
def test_crossref_query(  # type: ignore
    doi: str,
    expected_dict: dict,
) -> None:
    """Test the crossref query_doi()"""

    api = crossref_api.CrossrefAPI(params={})

    # replace the / in filenames by _
    filename = Path(__file__).parent / f"data/{doi.replace('/', '_')}.json"
    with open(filename, encoding="utf-8") as file:
        json_str = file.read()

    with requests_mock.Mocker() as req_mock:
        req_mock.get(
            f"https://api.crossref.org/works/{doi}", content=json_str.encode("utf-8")
        )

        actual = api.query_doi(doi=doi)
        expected = colrev.record.record_prep.PrepRecord(expected_dict)

        assert actual.data == expected.data



#! /usr/bin/env python
"""Convenience functions to write csv files"""
from __future__ import annotations

import pandas as pd

from colrev.constants import Fields

FIELDS = [
    Fields.ID,
    Fields.ENTRYTYPE,
    Fields.TITLE,
    Fields.AUTHOR,
    Fields.YEAR,
    Fields.JOURNAL,
    Fields.BOOKTITLE,
    Fields.VOLUME,
    Fields.NUMBER,
    Fields.PAGES,
    Fields.DOI,
    Fields.URL,
    Fields.FILE,
]


def to_dataframe(*, records_dict: dict) -> pd.DataFrame:
    """Convert a records dict to a pandas DataFrame"""
    data = []
    for record_id in sorted(records_dict.keys()):
        record_dict = records_dict[record_id]
        row = {}
        for field in FIELDS:
            if field in record_dict:
                row[field] = record_dict[field]
            else:
                row[field] = ""
        data.append(row)
    return pd.DataFrame(data)


def write_file(*, records_dict: dict, filename: str) -> None:
    """Write a csv file from a records dict"""
    data_frame = to_dataframe(records_dict=records_dict)
    data_frame.to_csv(filename, index=False)



#! /usr/bin/env python
"""Europe PMC API"""
import typing
from xml.etree.ElementTree import Element  # nosec

import requests
from lxml import etree

import colrev.env.language_service
import colrev.record.record_prep
from colrev.constants import ENTRYTYPES
from colrev.constants import Fields

# pylint: disable=too-few-public-methods


class EPMCAPI:
    """Connector for the Europe PMC API"""

    def __init__(self, params: dict, email: str, session: requests.Session) -> None:
        self.params = params

        self.url = (
            "https://www.ebi.ac.uk/europepmc/webservices/rest/search?query="
            + params["query"]
        )
        self.email = email
        self.session = session
        self.headers = {"user-agent": f"{__name__} (mailto:{email})"}

    def get_records(self) -> typing.Iterator[colrev.record.record_prep.PrepRecord]:
        """Get records from the Europe PMC API"""

        ret = self.session.request("GET", self.url, headers=self.headers, timeout=60)
        ret.raise_for_status()
        if ret.status_code != 200:
            # review_manager.logger.debug(
            #     f"europe_pmc failed with status {ret.status_code}"
            # )
            return

        root = etree.fromstring(str.encode(ret.text))
        result_list = root.findall("resultList")[0]
        for result_item in result_list.findall("result"):
            retrieved_record = self._europe_pmc_xml_to_record(item=result_item)
            yield retrieved_record

        self.url = ""
        next_page_url_node = root.find("nextPageUrl")
        if next_page_url_node is not None:
            if next_page_url_node.text is not None:
                self.url = next_page_url_node.text

    @classmethod
    def _get_string_from_item(cls, *, item, key: str) -> str:  # type: ignore
        return_string = ""
        for selected_node in item.findall(key):
            return_string = selected_node.text
        return return_string

    # pylint: disable=colrev-missed-constant-usage
    @classmethod
    def _europe_pmc_xml_to_record(
        cls, *, item: Element
    ) -> colrev.record.record_prep.PrepRecord:
        retrieved_record_dict: dict = {Fields.ENTRYTYPE: ENTRYTYPES.ARTICLE}
        retrieved_record_dict[Fields.AUTHOR] = cls._get_string_from_item(
            item=item, key="authorString"
        )
        retrieved_record_dict[Fields.JOURNAL] = cls._get_string_from_item(
            item=item, key="journalTitle"
        )
        retrieved_record_dict[Fields.DOI] = cls._get_string_from_item(
            item=item, key="doi"
        )
        retrieved_record_dict[Fields.TITLE] = cls._get_string_from_item(
            item=item, key="title"
        )
        retrieved_record_dict[Fields.YEAR] = cls._get_string_from_item(
            item=item, key="pubYear"
        )
        retrieved_record_dict[Fields.VOLUME] = cls._get_string_from_item(
            item=item, key="journalVolume"
        )
        retrieved_record_dict[Fields.NUMBER] = cls._get_string_from_item(
            item=item, key="issue"
        )
        retrieved_record_dict[Fields.PUBMED_ID] = cls._get_string_from_item(
            item=item, key="pmid"
        )
        retrieved_record_dict[Fields.PMCID] = cls._get_string_from_item(
            item=item, key="pmcid"
        )

        retrieved_record_dict["epmc_source"] = cls._get_string_from_item(
            item=item, key="source"
        )
        retrieved_record_dict["epmc_id"] = cls._get_string_from_item(
            item=item, key="id"
        )
        retrieved_record_dict[Fields.EUROPE_PMC_ID] = (
            retrieved_record_dict.get("epmc_source", "NO_SOURCE")
            + "/"
            + retrieved_record_dict.get("epmc_id", "NO_ID")
        )
        retrieved_record_dict[Fields.ID] = retrieved_record_dict[Fields.EUROPE_PMC_ID]

        retrieved_record_dict = {
            k: v
            for k, v in retrieved_record_dict.items()
            if k not in ["epmc_id", "epmc_source"] and v != ""
        }

        record = colrev.record.record_prep.PrepRecord(retrieved_record_dict)
        return record



#! /usr/bin/env python
"""Checker for page-range."""
from __future__ import annotations

import re

import colrev.record.qm.quality_model
from colrev.constants import DefectCodes
from colrev.constants import Fields

# pylint: disable=too-few-public-methods


class PageRangeChecker:
    """The PageRangeChecker"""

    msg = DefectCodes.PAGE_RANGE

    def __init__(
        self, quality_model: colrev.record.qm.quality_model.QualityModel
    ) -> None:
        self.quality_model = quality_model

    def run(self, *, record: colrev.record.record.Record) -> None:
        """Run the page-range checks"""

        if (
            Fields.PAGES not in record.data
            or record.ignored_defect(key=Fields.PAGES, defect=self.msg)
            or not re.match(r"^\d+\-\-\d+$", record.data[Fields.PAGES])
        ):
            return
        if record.masterdata_is_curated():
            return

        if self._pages_descending(pages=record.data[Fields.PAGES]):
            record.add_field_provenance_note(key=Fields.PAGES, note=self.msg)
        else:
            record.remove_field_provenance_note(key=Fields.PAGES, note=self.msg)

    def _pages_descending(self, *, pages: str) -> bool:
        from_page, to_page = re.findall(r"(\d+)", pages)
        if int(from_page) > int(to_page):
            return True

        return False


def register(quality_model: colrev.record.qm.quality_model.QualityModel) -> None:
    """Register the checker"""
    quality_model.register_checker(PageRangeChecker(quality_model))



#! /usr/bin/env python
"""Consolidation of metadata based on SemanticScholar API as a prep operation"""
from __future__ import annotations

import json

import requests
import zope.interface
from pydantic import Field

import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
import colrev.record.record_prep
import colrev.record.record_similarity
from colrev.constants import Fields
from colrev.packages.semanticscholar.src import record_transformer

# pylint: disable=too-few-public-methods
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.package_manager.interfaces.PrepInterface)
class SemanticScholarPrep:
    """Prepares records based on SemanticScholar metadata"""

    settings_class = colrev.package_manager.package_settings.DefaultSettings
    ci_supported: bool = Field(default=True)

    source_correction_hint = (
        "fill out the online form: "
        + "https://www.semanticscholar.org/faq#correct-error"
    )
    always_apply_changes = False

    def __init__(
        self,
        *,
        prep_operation: colrev.ops.prep.Prep,
        settings: dict,
    ) -> None:
        self.settings = self.settings_class(**settings)
        self.prep_operation = prep_operation
        self.review_manager = prep_operation.review_manager
        _, email = prep_operation.review_manager.get_committer()
        self.headers = {"user-agent": f"{__name__} (mailto:{email})"}
        self.session = prep_operation.review_manager.get_cached_session()

    def _retrieve_record_from_semantic_scholar(
        self,
        record_in: colrev.record.record_prep.PrepRecord,
    ) -> colrev.record.record_prep.PrepRecord:
        """Prepare the record metadata based on SemanticScholar"""

        search_api_url = "https://api.semanticscholar.org/graph/v1/paper/search?query="
        url = search_api_url + record_in.data.get(Fields.TITLE, "").replace(" ", "+")

        # prep_operation.review_manager.logger.debug(url)
        ret = self.session.request(
            "GET", url, headers=self.headers, timeout=self.prep_operation.timeout
        )
        ret.raise_for_status()

        data = json.loads(ret.text)
        items = data["data"]
        if len(items) == 0:
            return record_in
        if "paperId" not in items[0]:
            return record_in

        paper_id = items[0]["paperId"]
        record_retrieval_url = "https://api.semanticscholar.org/v1/paper/" + paper_id
        # prep_operation.review_manager.logger.debug(record_retrieval_url)
        ret_ent = self.session.request(
            "GET",
            record_retrieval_url,
            headers=self.headers,
            timeout=self.prep_operation.timeout,
        )
        ret_ent.raise_for_status()
        item = json.loads(ret_ent.text)

        retrieved_record = record_transformer.dict_to_record(item=item)
        retrieved_record.add_provenance_all(source=record_retrieval_url)

        return retrieved_record.copy_prep_rec()

    def prepare(
        self, record: colrev.record.record_prep.PrepRecord
    ) -> colrev.record.record.Record:
        """Prepare a record based on metadata from SemanticScholar"""

        try:
            retrieved_record = self._retrieve_record_from_semantic_scholar(record)
            if Fields.SEMANTIC_SCHOLAR_ID not in retrieved_record.data:
                return record

            # Remove fields that are not/rarely available before
            # calculating similarity metrics
            orig_record = record.copy_prep_rec()
            for key in [Fields.VOLUME, Fields.NUMBER, Fields.NUMBER, Fields.PAGES]:
                if key in orig_record.data:
                    record.remove_field(key=key)

            if not colrev.record.record_similarity.matches(record, retrieved_record):
                return record

            record.merge(
                retrieved_record,
                default_source=retrieved_record.data[Fields.SEMANTIC_SCHOLAR_ID],
            )

        except (requests.exceptions.RequestException,):
            pass
        return record



#! /usr/bin/env python3
"""CoLRev dashboard operation: track project progress through dashboard"""
from __future__ import annotations

import dash  # pylint: disable=import-error
from dash import Dash  # pylint: disable=import-error
from dash import html  # pylint: disable=import-error

# dash dependencies optional
# install with pip install colrev[ui_web]
# will fail if not installed


# pylint: disable=too-few-public-methods


class Dashboard:
    """Dashboard class"""

    def make_dashboard(self) -> Dash:
        """creates dashboard header and general structure"""
        app = Dash(__name__, use_pages=True)

        app.layout = html.Div(
            [
                html.Div(
                    [
                        html.Img(src="assets/favicon.ico", className="logo"),
                        html.H1(children="Dashboard", className="header-title"),
                        html.H2(
                            children="make progress visible",
                            className="header-subtitle",
                        ),
                    ],
                    className="header",
                ),
                dash.page_container,
            ]
        )

        return app


def main() -> None:
    """Main method for the dashboard"""

    dashboard = Dashboard()
    app = dashboard.make_dashboard()
    app.run_server()  # debug=True



#! /usr/bin/env python
"""Pubmed API"""
import datetime
import logging
import time
import typing
from sqlite3 import OperationalError
from xml.etree import ElementTree  # nosec
from xml.etree.ElementTree import Element  # nosec

import requests
from lxml import etree
from lxml import html
from lxml.etree import XMLSyntaxError

import colrev.exceptions as colrev_exceptions
import colrev.record.record_prep
from colrev.constants import Fields


# pylint: disable=too-few-public-methods


class PubmedAPI:
    """Connector for the Pubmed API"""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        *,
        parameters: dict,
        email: str,
        session: requests.Session,
        timeout: int = 60,
        logger: typing.Optional[logging.Logger] = None,
    ):
        self.email = email
        self.session = session
        self._timeout = timeout
        self.params = parameters

        self.headers = {"user-agent": f"{__name__} (mailto:{self.email})"}

        if logger is not None:
            self.logger = logger
        else:
            self.logger = logging.getLogger(__name__)

    @classmethod
    def _get_author_string_from_node(cls, *, author_node: Element) -> str:
        authors_string = ""
        author_last_name_node = author_node.find("LastName")
        if author_last_name_node is not None:
            if author_last_name_node.text is not None:
                authors_string += author_last_name_node.text
        author_fore_name_node = author_node.find("ForeName")
        if author_fore_name_node is not None:
            if author_fore_name_node.text is not None:
                authors_string += ", "
                authors_string += author_fore_name_node.text
        return authors_string

    @classmethod
    def _get_author_string(cls, *, root) -> str:  # type: ignore
        authors_list = []
        for author_node in root.xpath(
            "/PubmedArticleSet/PubmedArticle/MedlineCitation/Article/AuthorList/Author"
        ):
            authors_list.append(
                cls._get_author_string_from_node(author_node=author_node)
            )
        return " and ".join(authors_list)

    @classmethod
    def _get_title_string(cls, *, root) -> str:  # type: ignore
        title = root.xpath(
            "/PubmedArticleSet/PubmedArticle/MedlineCitation/Article/ArticleTitle"
        )
        if title:
            if title[0].text:
                title = title[0].text.strip().rstrip(".")
                if title.startswith("[") and title.endswith("]"):
                    title = title[1:-1]
                return title
        return ""

    @classmethod
    def _get_abstract_string(cls, *, root) -> str:  # type: ignore
        abstract = root.xpath(
            "/PubmedArticleSet/PubmedArticle/MedlineCitation/Article/Abstract"
        )
        if abstract:
            return ElementTree.tostring(abstract[0], encoding="unicode")
        return ""

    # pylint: disable=colrev-missed-constant-usage
    @classmethod
    def _pubmed_xml_to_record(cls, *, root) -> dict:  # type: ignore
        retrieved_record_dict: dict = {Fields.ENTRYTYPE: "misc"}

        pubmed_article = root.find("PubmedArticle")
        if pubmed_article is None:
            return {}
        if pubmed_article.find("MedlineCitation") is None:
            return {}

        retrieved_record_dict[Fields.TITLE] = cls._get_title_string(root=root)
        retrieved_record_dict[Fields.AUTHOR] = cls._get_author_string(root=root)

        journal_path = "/PubmedArticleSet/PubmedArticle/MedlineCitation/Article/Journal"
        journal_name = root.xpath(journal_path + "/ISOAbbreviation")
        if journal_name:
            retrieved_record_dict[Fields.ENTRYTYPE] = "article"
            retrieved_record_dict[Fields.JOURNAL] = journal_name[0].text

        volume = root.xpath(journal_path + "/JournalIssue/Volume")
        if volume:
            retrieved_record_dict[Fields.VOLUME] = volume[0].text

        number = root.xpath(journal_path + "/JournalIssue/Issue")
        if number:
            retrieved_record_dict[Fields.NUMBER] = number[0].text

        year = root.xpath(journal_path + "/JournalIssue/PubDate/Year")
        if year:
            retrieved_record_dict[Fields.YEAR] = year[0].text

        retrieved_record_dict[Fields.ABSTRACT] = cls._get_abstract_string(root=root)

        article_id_list = root.xpath(
            "/PubmedArticleSet/PubmedArticle/PubmedData/ArticleIdList"
        )
        for article_id in article_id_list[0]:
            id_type = article_id.attrib.get("IdType")
            if article_id.attrib.get("IdType") == "pubmed":
                retrieved_record_dict["pubmedid"] = article_id.text.upper()
            elif article_id.attrib.get("IdType") == "doi":
                retrieved_record_dict[Fields.DOI] = article_id.text.upper()
            else:
                retrieved_record_dict[id_type] = article_id.text

        retrieved_record_dict = {
            k: v for k, v in retrieved_record_dict.items() if v != ""
        }
        if (
            retrieved_record_dict.get("pii", "pii").lower()
            == retrieved_record_dict.get("doi", "doi").lower()
        ):
            del retrieved_record_dict["pii"]

        return retrieved_record_dict

    def query_id(self, *, pubmed_id: str) -> colrev.record.record.Record:
        """Retrieve records from Pubmed based on a query"""

        try:
            database = "pubmed"
            url = (
                "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?"
                + f"db={database}&id={pubmed_id}&rettype=xml&retmode=text"
            )

            while True:
                # review_manager.logger.debug(url)
                ret = self.session.request(
                    "GET", url, headers=self.headers, timeout=self._timeout
                )
                if ret.status_code == 429:
                    time.sleep(10)
                    continue
                ret.raise_for_status()
                if ret.status_code != 200:
                    # review_manager.logger.debug(
                    #     f"crossref_query failed with status {ret.status_code}"
                    # )
                    raise colrev_exceptions.SearchSourceException(
                        "Pubmed record not found"
                    )

                root = etree.fromstring(str.encode(ret.text))
                retrieved_record_dict = self._pubmed_xml_to_record(root=root)
                if not retrieved_record_dict:
                    raise colrev_exceptions.SearchSourceException(
                        "Pubmed record not found"
                    )
                retrieved_record = colrev.record.record.Record(retrieved_record_dict)
                return retrieved_record
        except requests.exceptions.RequestException as exc:
            raise colrev_exceptions.SearchSourceException(
                "Pubmed record not found"
            ) from exc
        except XMLSyntaxError as exc:
            raise colrev_exceptions.RecordNotParsableException(
                "Error parsing xml"
            ) from exc
        # pylint: disable=duplicate-code
        except OperationalError as exc:
            raise colrev_exceptions.ServiceNotAvailableException(
                "sqlite, required for requests CachedSession "
                "(possibly caused by concurrent operations)"
            ) from exc

    def _get_pubmed_ids(self, query: str, retstart: int, page: int) -> dict:

        if not query.startswith("https://pubmed.ncbi.nlm.nih.gov/?term="):
            query = "https://pubmed.ncbi.nlm.nih.gov/?term=" + query
        url = query + f"&retstart={retstart}&page={page}"
        ret = self.session.request("GET", url, headers=self.headers, timeout=30)
        ret.raise_for_status()
        # if ret.status_code != 200:
        #     # review_manager.logger.debug(
        #     #     f"crossref_query failed with status {ret.status_code}"
        #     # )
        #     return []

        root = html.fromstring(str.encode(ret.text))
        meta_tags = root.findall(".//meta[@name='log_displayeduids']")
        displayed_uids = [tag.get("content") for tag in meta_tags][0].split(",")

        total_results_items = root.xpath(
            "//div[@class='results-amount']//span[@class='value']/text()"
        )
        total_results = int(total_results_items[0].replace(",", ""))
        # total_results = [int(result.replace(',', '')) for result in total_results]

        return {"uids": displayed_uids, "totalResults": total_results}

    def get_query_return(self) -> typing.Iterator[colrev.record.record.Record]:
        """Retrieve records from Pubmed based on a query"""

        retstart = 10
        page = 1
        while True:
            ret = self._get_pubmed_ids(
                query=self.params["query"], retstart=retstart, page=page
            )
            if page == 1:
                # pylint: disable=logging-fstring-interpolation
                total_results = ret["totalResults"]
                self.logger.info(f"Total results: {total_results}")
                expected_time = datetime.timedelta(seconds=total_results / 3)
                expected_time_seconds = round(expected_time.total_seconds())
                hours, remainder = divmod(expected_time_seconds, 3600)
                minutes, seconds = divmod(remainder, 60)
                expected_time_formatted = f"{hours:02d}:{minutes:02d}:{seconds:02d}"
                self.logger.info("Rate limit: 3 requests per second")
                self.logger.info(f"Expected time [hh:mm:ss]: {expected_time_formatted}")

            pubmed_ids = ret["uids"]
            if not pubmed_ids:
                break
            for pubmed_id in pubmed_ids:
                time.sleep(0.5)  # to avoid 429 error (pubmed allows 3rps)
                yield self.query_id(pubmed_id=pubmed_id)

            page += 1



#! /usr/bin/env python
"""Prescreen based on a table"""
from __future__ import annotations

import csv
from pathlib import Path

import pandas as pd
import zope.interface
from pydantic import Field

import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import Colors
from colrev.constants import Fields
from colrev.constants import RecordState


# pylint: disable=too-few-public-methods
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.package_manager.interfaces.PrescreenInterface)
class TablePrescreen:
    """Table-based prescreen (exported and imported)"""

    settings_class = colrev.package_manager.package_settings.DefaultSettings
    ci_supported: bool = Field(default=False)
    export_todos_only: bool = True

    def __init__(
        self,
        *,
        prescreen_operation: colrev.ops.prescreen.Prescreen,
        settings: dict,
    ) -> None:
        self.review_manager = prescreen_operation.review_manager
        self.settings = self.settings_class(**settings)

    def export_table(
        self,
        *,
        records: dict,
        split: list,
        export_table_format: str = "csv",
    ) -> None:
        """Export a prescreen table"""

        # gh_issue https://github.com/CoLRev-Environment/colrev/issues/73
        # add delta (records not yet in the table)
        # instead of overwriting
        # export_table_format as a settings parameter

        self.review_manager.logger.info("Loading records for export")

        tbl = []
        for record in records.values():
            if record[Fields.STATUS] not in [
                RecordState.md_processed,
                RecordState.rev_prescreen_excluded,
                RecordState.rev_prescreen_included,
                RecordState.pdf_needs_manual_retrieval,
                RecordState.pdf_imported,
                RecordState.pdf_not_available,
                RecordState.pdf_needs_manual_preparation,
                RecordState.pdf_prepared,
                RecordState.rev_excluded,
                RecordState.rev_included,
                RecordState.rev_synthesized,
            ]:
                continue

            if len(split) > 0:
                if record[Fields.ID] not in split:
                    continue

            if RecordState.md_processed == record[Fields.STATUS]:
                inclusion_1 = "TODO"
            elif self.export_todos_only:
                continue
            elif RecordState.rev_prescreen_excluded == record[Fields.STATUS]:
                inclusion_1 = "out"
            else:
                inclusion_1 = "in"

            # pylint: disable=duplicate-code
            row = {
                Fields.ID: record[Fields.ID],
                Fields.AUTHOR: record.get(Fields.AUTHOR, ""),
                Fields.TITLE: record.get(Fields.TITLE, ""),
                Fields.JOURNAL: record.get(Fields.JOURNAL, ""),
                Fields.BOOKTITLE: record.get(Fields.BOOKTITLE, ""),
                Fields.YEAR: record.get(Fields.YEAR, ""),
                Fields.VOLUME: record.get(Fields.VOLUME, ""),
                Fields.NUMBER: record.get(Fields.NUMBER, ""),
                Fields.PAGES: record.get(Fields.PAGES, ""),
                Fields.DOI: record.get(Fields.DOI, ""),
                Fields.ABSTRACT: record.get(Fields.ABSTRACT, ""),
                "presceen_inclusion": inclusion_1,
            }
            tbl.append(row)

        if export_table_format.lower() == "csv":
            screen_df = pd.DataFrame(tbl)
            screen_df.to_csv("prescreen.csv", index=False, quoting=csv.QUOTE_ALL)
            self.review_manager.logger.info("Created prescreen.csv")

        if export_table_format.lower() == "xlsx":
            screen_df = pd.DataFrame(tbl)
            screen_df.to_excel("prescreen.xlsx", index=False, sheet_name="screen")
            self.review_manager.logger.info("Created prescreen.xlsx")

        self.review_manager.logger.info(
            f"To prescreen records, {Colors.ORANGE}enter [in|out] "
            f"in the presceen_inclusion column.{Colors.END}"
        )
        self.review_manager.logger.info(
            f"Afterwards, run {Colors.ORANGE}colrev prescreen --import_table "
            f"prescreen.{export_table_format.lower()}{Colors.END}"
        )

    def import_table(
        self,
        *,
        records: dict,
        import_table_path: str = "prescreen.csv",
    ) -> None:
        """Import a prescreen table"""

        # pylint: disable=too-many-branches

        self.review_manager.logger.info(f"Load {import_table_path}")

        # pylint: disable=duplicate-code
        if not Path(import_table_path).is_file():
            self.review_manager.logger.error(
                f"Did not find {import_table_path} - exiting."
            )
            return

        if import_table_path.endswith(".csv"):
            prescreen_df = pd.read_csv(import_table_path)
        elif import_table_path.endswith(".xlsx") or import_table_path.endswith(".xls"):
            prescreen_df = pd.read_excel(import_table_path)
        else:
            raise ValueError(f"Unsupported file format: {import_table_path}")
        prescreen_df.fillna("", inplace=True)
        prescreened_records = prescreen_df.to_dict("records")

        if "presceen_inclusion" not in prescreened_records[0]:
            self.review_manager.logger.warning("presceen_inclusion column missing")
            return

        prescreen_included = 0
        prescreen_excluded = 0
        nr_todo = 0
        self.review_manager.logger.info("Update prescreen results")
        for prescreened_record in prescreened_records:
            if prescreened_record.get(Fields.ID, "") in records:
                record = colrev.record.record.Record(
                    records[prescreened_record.get(Fields.ID, "")]
                )
                if record.data[Fields.STATUS] in RecordState.get_post_x_states(
                    state=RecordState.rev_prescreen_included
                ):
                    if (
                        "in" == prescreened_record.data.get("presceen_inclusion", "")
                        and RecordState.rev_prescreen_excluded
                        != record.data[Fields.STATUS]
                    ):
                        continue

                if prescreened_record.get("presceen_inclusion", "") == "out":
                    if record.data[Fields.STATUS] != RecordState.rev_prescreen_excluded:
                        prescreen_excluded += 1
                    record.set_status(RecordState.rev_prescreen_excluded)

                elif prescreened_record.get("presceen_inclusion", "") == "in":
                    if record.data[Fields.STATUS] != RecordState.rev_prescreen_included:
                        prescreen_included += 1
                    record.set_status(RecordState.rev_prescreen_included)
                elif prescreened_record.get("presceen_inclusion", "") == "TODO":
                    nr_todo += 1
                else:
                    self.review_manager.logger.warning(
                        "Invalid value in prescreen_inclusion: "
                        f"{prescreened_record.get('presceen_inclusion', '')} "
                        f"({prescreened_record.get('ID', 'NO_ID')})"
                    )

            else:
                self.review_manager.logger.warning(
                    f"ID not in records: {prescreened_record.get('ID', '')}"
                )

        self.review_manager.logger.info(
            f" {Colors.GREEN}{prescreen_included} records prescreen_included{Colors.END}"
        )
        self.review_manager.logger.info(
            f" {Colors.RED}{prescreen_excluded} records prescreen_excluded{Colors.END}"
        )

        self.review_manager.logger.info(
            f" {Colors.ORANGE}{nr_todo} records to prescreen{Colors.END}"
        )

        self.review_manager.dataset.save_records_dict(records)
        self.review_manager.logger.info("Completed import")

    def run_prescreen(
        self,
        records: dict,
        split: list,
    ) -> dict:
        """Prescreen records based on screening tables"""

        if input("create prescreen table [y,n]?") == "y":
            self.export_table(records=records, split=split)

        if input("import prescreen table [y,n]?") == "y":
            self.import_table(records=records)

        if self.review_manager.dataset.has_record_changes():
            if input("create commit [y,n]?") == "y":
                self.review_manager.dataset.create_commit(
                    msg="Pre-screen (table)",
                    manual_author=True,
                )
        return records



#! /usr/bin/env python
"""Dedupe functionality dedicated to curated metadata repositories"""
from __future__ import annotations

import numpy as np
import pandas as pd
import zope.interface
from pydantic import BaseModel
from pydantic import Field
from rapidfuzz import fuzz
from tqdm import tqdm

import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import Colors
from colrev.constants import Fields
from colrev.constants import RecordState


# pylint: disable=too-many-arguments
# pylint: disable=too-few-public-methods
# pylint: disable=duplicate-code


class CurationDedupeSettings(BaseModel):
    """Settings for CurationDedupe"""

    endpoint: str
    selected_source: str


@zope.interface.implementer(colrev.package_manager.interfaces.DedupeInterface)
class CurationDedupe:
    """Deduplication endpoint for curations with full journals/proceedings
    retrieved from different sources (identifying duplicates in groups of
    volumes/issues or years)"""

    settings_class = CurationDedupeSettings
    settings: CurationDedupeSettings
    ci_supported: bool = Field(default=True)

    def __init__(
        self,
        *,
        dedupe_operation: colrev.ops.dedupe.Dedupe,
        settings: dict,
    ):
        self.settings = self.settings_class(**settings)
        self.dedupe_operation = dedupe_operation
        self.review_manager = dedupe_operation.review_manager

        self.pdf_qm = self.review_manager.get_pdf_qm()

    def _has_overlapping_colrev_id(
        self,
        record_a: colrev.record.record.Record,
        record_b: colrev.record.record.Record,
    ) -> bool:
        """Check if a record has an overlapping colrev_id with the other record"""
        own_colrev_ids = record_a.get_colrev_id()
        other_colrev_ids = record_b.get_colrev_id()
        if len(own_colrev_ids) > 0 and len(other_colrev_ids) > 0:
            if any(cid in own_colrev_ids for cid in other_colrev_ids):
                return True
        return False

    def _get_similarity(self, *, df_a: pd.Series, df_b: pd.Series) -> float:
        author_similarity = fuzz.ratio(df_a[Fields.AUTHOR], df_b[Fields.AUTHOR]) / 100

        title_similarity = (
            fuzz.ratio(df_a[Fields.TITLE].lower(), df_b[Fields.TITLE].lower()) / 100
        )

        # Note : the toc-based processing means that we are robust against
        # outlet, year, volume, number variations!
        weights = [0.4, 0.6]
        similarities = [
            author_similarity,
            title_similarity,
        ]

        weighted_average = sum(
            similarities[g] * weights[g] for g in range(len(similarities))
        )

        similarity_score = round(weighted_average, 4)

        return similarity_score

    def _calculate_similarities(
        self,
        *,
        similarity_array: np.ndarray,
        references: pd.DataFrame,
        min_similarity: float,
    ) -> tuple:
        # Fill out the similarity matrix first
        for base_entry_i in range(1, references.shape[0]):
            for comparison_entry_i in range(1, references.shape[0]):
                if base_entry_i > comparison_entry_i:
                    if -1 != similarity_array[base_entry_i, comparison_entry_i]:
                        similarity_array[base_entry_i, comparison_entry_i] = (
                            self._get_similarity(
                                df_a=references.iloc[base_entry_i],
                                df_b=references.iloc[comparison_entry_i],
                            )
                        )

        tuples_to_process = []
        maximum_similarity = 1
        while True:
            maximum_similarity = np.amax(similarity_array)
            if maximum_similarity < min_similarity:
                break
            result = np.where(similarity_array == np.amax(similarity_array))
            list_of_coordinates = list(zip(result[0], result[1]))
            for cord in list_of_coordinates:
                similarity_array[cord] = 0  # ie., has been processed
                tuples_to_process.append(
                    [
                        references.iloc[cord[0]][Fields.ID],
                        references.iloc[cord[1]][Fields.ID],
                        maximum_similarity,
                        "not_processed",
                    ]
                )

        return similarity_array, tuples_to_process

    def _get_toc_items(self, *, records_list: list) -> list:
        toc_items = []
        for record in records_list:
            toc_item = {}
            if record[Fields.ENTRYTYPE] == "article":
                if Fields.JOURNAL in record:
                    toc_item[Fields.JOURNAL] = record[Fields.JOURNAL]
                if Fields.VOLUME in record:
                    toc_item[Fields.VOLUME] = record[Fields.VOLUME]
                if Fields.NUMBER in record:
                    toc_item[Fields.NUMBER] = record[Fields.NUMBER]

            if record[Fields.ENTRYTYPE] == "inproceedings":
                if Fields.BOOKTITLE in record:
                    toc_item[Fields.BOOKTITLE] = record[Fields.BOOKTITLE]
                    toc_item[Fields.YEAR] = record[Fields.YEAR]
            if len(toc_item) > 0:
                toc_items.append(toc_item)

        temp = {tuple(sorted(sub.items())) for sub in toc_items}
        toc_items = list(map(dict, temp))  # type: ignore
        return toc_items

    def _warn_on_missing_sources(self, *, first_source: bool) -> None:
        # warn if not all SOURCE.filenames are included in a dedupe script
        if first_source:
            available_sources = [
                str(s.filename)
                for s in self.review_manager.settings.sources
                if "md_" not in str(s.filename)
            ]
            dedupe_sources = [
                s["selected_source"]
                for s in self.review_manager.settings.dedupe.dedupe_package_endpoints
                if "colrev.curation_full_outlet_dedupe" == s["endpoint"]
            ]
            sources_missing_in_dedupe = [
                x for x in available_sources if x not in dedupe_sources
            ]
            if len(sources_missing_in_dedupe) > 0:
                self.review_manager.logger.warning(
                    f"{Colors.ORANGE}Sources missing in "
                    "dedupe.scripts.colrev.curation_full_outlet_dedupe: "
                    f"{','.join(sources_missing_in_dedupe)}{Colors.END}"
                )
                if input("Add sources [y,n]?") == "y":
                    for source_missing_in_dedupe in sources_missing_in_dedupe:
                        dedupe_package_endpoints = (
                            self.review_manager.settings.dedupe.dedupe_package_endpoints
                        )
                        penultimate_position = len(dedupe_package_endpoints) - 1
                        dedupe_script_to_add = {
                            "endpoint": "colrev.curation_full_outlet_dedupe",
                            "selected_source": source_missing_in_dedupe,
                        }
                        dedupe_package_endpoints.insert(
                            penultimate_position, dedupe_script_to_add
                        )
                        self.review_manager.save_settings()
                        self.review_manager.logger.info(
                            f"{Colors.GREEN}Added {source_missing_in_dedupe} "
                            f"to dedupe.scripts{Colors.END}"
                        )

    def _add_first_source_if_deduplicated(self, *, records: dict) -> None:
        self.review_manager.logger.info(
            f"Starting with records from {self.settings.selected_source}"
            " (setting to md_processed as the initial records)"
        )

        source_records = [
            r
            for r in records.values()
            if r[Fields.STATUS] == RecordState.md_prepared
            and any(
                self.settings.selected_source.replace("data/search/", "") in co
                for co in r[Fields.ORIGIN]
            )
        ]

        toc_items = self._get_toc_items(records_list=source_records)

        for toc_item in toc_items:
            # Note : these would be potential errors (duplicates)
            # because they have the same selected_source
            processed_same_toc_same_source_records = [
                r
                for r in records.values()
                if all(r.get(k, "NA") == v for k, v in toc_item.items())
                and r[Fields.STATUS]
                not in [
                    RecordState.md_prepared,
                    RecordState.md_needs_manual_preparation,
                    RecordState.md_imported,
                    RecordState.rev_prescreen_excluded,
                ]
                and any(
                    self.settings.selected_source.replace("data/search/", "") in co
                    for co in r[Fields.ORIGIN]
                )
            ]
            if 0 == len(processed_same_toc_same_source_records):
                print("\n\n")
                print(toc_item)

                for source_record_dict in sorted(
                    source_records, key=lambda d: d[Fields.AUTHOR]
                ):
                    if all(
                        source_record_dict.get(k, "NA") == v
                        for k, v in toc_item.items()
                    ):
                        # Record(sr).print_citation_format()
                        print(
                            f"{source_record_dict.get('author', 'NO_AUTHOR')} : "
                            f"{source_record_dict.get('title', 'NO_TITLE')}"
                        )
                recs_unique = self.review_manager.force_mode
                if not recs_unique:
                    recs_unique = "y" == input(
                        "No existing records (md_processed*) found. "
                        "All records unique? Set to md_processed [y]? "
                    )
                if recs_unique:
                    for source_record_dict in source_records:
                        if all(
                            source_record_dict.get(k, "NA") == v
                            for k, v in toc_item.items()
                        ):
                            source_record = colrev.record.record.Record(
                                data=source_record_dict
                            )
                            source_record.set_status(
                                target_state=RecordState.md_processed
                            )
            else:
                print(toc_item)
                print("Pre-imported records found for this toc_item (skipping)")
                # print(processed_same_toc_same_source_records)

        for record in records.values():
            record.pop(Fields.CONTAINER_TITLE)
        self.review_manager.dataset.save_records_dict(records)

        if self.review_manager.dataset.has_record_changes():
            self.review_manager.logger.info(f"{Colors.GREEN}Commit changes{Colors.END}")
            self.review_manager.dataset.create_commit(
                msg=(
                    "Merge duplicate records (set unique records from "
                    f"{self.settings.selected_source} "
                    "to md_processed)"
                ),
            )
        else:
            self.review_manager.logger.info(
                f"{Colors.GREEN}No duplicates found{Colors.END}"
            )

    def _prep_records(self, *, records: dict) -> dict:
        required_fields = [
            Fields.TITLE,
            Fields.AUTHOR,
            Fields.YEAR,
            Fields.JOURNAL,
            Fields.VOLUME,
            Fields.NUMBER,
            Fields.PAGES,
            Fields.BOOKTITLE,
        ]
        for record in records.values():
            if Fields.CONTAINER_TITLE not in record:
                record[Fields.CONTAINER_TITLE] = (
                    record.get(Fields.JOURNAL, "")
                    + record.get(Fields.BOOKTITLE, "")
                    + record.get(Fields.SERIES, "")
                )

            for required_field in required_fields:
                if required_field not in record:
                    record[required_field] = ""
        return records

    def _dedupe_source(self, *, records: dict) -> list[list]:
        self.review_manager.logger.info(
            "Processing as a non-pdf source (matching exact colrev_ids)"
        )

        source_records = [
            r
            for r in records.values()
            if r[Fields.STATUS] == RecordState.md_prepared
            and any(
                self.settings.selected_source.replace("data/search/", "") in co
                for co in r[Fields.ORIGIN]
            )
        ]

        toc_items = self._get_toc_items(records_list=source_records)

        decision_list: list[list] = []
        # decision_list =[{'ID1': ID1, 'ID2': ID2, 'decision': 'duplicate'}]

        # match based on overlapping  colrev_ids
        for toc_item in tqdm(toc_items):
            processed_same_toc_records = [
                r
                for r in records.values()
                if all(r.get(k, "NA") == v for k, v in toc_item.items())
                and r[Fields.STATUS]
                not in [
                    RecordState.md_imported,
                    RecordState.md_needs_manual_preparation,
                    RecordState.md_prepared,
                    RecordState.rev_prescreen_excluded,
                ]
                and not any(
                    self.settings.selected_source.replace("data/search/", "") in co
                    for co in r[Fields.ORIGIN]
                )
            ]
            new_same_toc_records = [
                r
                for r in source_records
                if all(r.get(k, "NA") == v for k, v in toc_item.items())
            ]
            if len(new_same_toc_records) > 0:
                # print(new_same_toc_records)
                for new_same_toc_record in new_same_toc_records:
                    for rec2 in processed_same_toc_records:
                        overlapping_colrev_ids = self._has_overlapping_colrev_id(
                            colrev.record.record.Record(new_same_toc_record),
                            colrev.record.record.Record(rec2),
                        )
                        if overlapping_colrev_ids:
                            decision_list.append(
                                [new_same_toc_record[Fields.ID], rec2[Fields.ID]]
                            )
                            print("TODO : validate whether it merges correctly:")
                            input(decision_list)

        return decision_list

    def _validate_potential_merge(self, *, rec1: dict, rec2: dict) -> bool:
        if Fields.FILE in rec2:
            updated_record_dict = rec1.copy()
            updated_record_dict[Fields.FILE] = rec2[Fields.FILE]
        elif Fields.FILE in rec1:
            updated_record_dict = rec2.copy()
            updated_record_dict[Fields.FILE] = rec1[Fields.FILE]
        else:  # None of the records is curated
            raise FileNotFoundError

        updated_record = colrev.record.record.Record(updated_record_dict)
        updated_record.run_pdf_quality_model(self.pdf_qm)
        return updated_record.has_pdf_defects()

    def _process_pdf_tuple(
        self,
        *,
        tuple_to_process: tuple,
        records: dict,
        decision_list: list,
        curated_record_ids: list,
        pdf_record_ids: list,
    ) -> None:
        rec1 = records[tuple_to_process[0]]
        rec2 = records[tuple_to_process[1]]

        # Note : Focus on merges between
        # curated_records and pdf_same_toc_records
        # Note : this should also ensure that pdf groups are not merged
        # until a corresponding curated record group is available.
        if (
            rec1[Fields.ID] in curated_record_ids
            and rec2[Fields.ID] in curated_record_ids
        ) or (rec1[Fields.ID] in pdf_record_ids and rec2[Fields.ID] in pdf_record_ids):
            return

        try:
            validated = self._validate_potential_merge(rec1=rec1, rec2=rec2)
        except FileNotFoundError:
            return

        overlapping_colrev_ids = self._has_overlapping_colrev_id(
            colrev.record.record.Record(rec1), colrev.record.record.Record(rec2)
        )
        if validated or overlapping_colrev_ids:
            decision_list.append([rec1[Fields.ID], rec2[Fields.ID]])

    def _dedupe_pdf_toc_item(
        self,
        *,
        decision_list: list[list],
        toc_item: dict,
        records: dict,
        source_records: list,
    ) -> None:
        processed_same_toc_records = [
            r
            for r in records.values()
            if all(r.get(k, "NA") == v for k, v in toc_item.items())
            and r[Fields.STATUS]
            not in [
                RecordState.md_imported,
                RecordState.md_needs_manual_preparation,
                RecordState.md_prepared,
                RecordState.rev_prescreen_excluded,
            ]
            and not any(
                self.settings.selected_source.replace("data/search/", "") in co
                for co in r[Fields.ORIGIN]
            )
        ]
        pdf_same_toc_records = [
            r
            for r in source_records
            if all(r.get(k, "NA") == v for k, v in toc_item.items())
        ]

        references = pd.DataFrame.from_records(
            processed_same_toc_records + pdf_same_toc_records
        )

        nr_entries = references.shape[0]
        if nr_entries == 0:
            return
        similarity_array = np.zeros([nr_entries, nr_entries])

        # Note : min_similarity only means that the PDF will be considered
        # for validates_based_on_metadata(...), which is the acutal test!
        similarity_array, tuples_to_process = self._calculate_similarities(
            similarity_array=similarity_array,
            references=references,
            min_similarity=0.7,
        )

        curated_record_ids = [r[Fields.ID] for r in processed_same_toc_records]
        pdf_record_ids = [r[Fields.ID] for r in pdf_same_toc_records]
        for tuple_to_process in tuples_to_process:
            self._process_pdf_tuple(
                tuple_to_process=tuple_to_process,
                records=records,
                decision_list=decision_list,
                curated_record_ids=curated_record_ids,
                pdf_record_ids=pdf_record_ids,
            )

    def _dedupe_pdf_source(self, *, records: dict) -> list[list]:
        self.review_manager.logger.info("Processing as a pdf source")

        source_records = [
            r
            for r in records.values()
            if r[Fields.STATUS] == RecordState.md_prepared
            and any(
                self.settings.selected_source.replace("data/search/", "") in co
                for co in r[Fields.ORIGIN]
            )
        ]

        decision_list: list[list] = []
        # decision_list =[{'ID1': ID1, 'ID2': ID2, 'decision': 'duplicate'}]

        for toc_item in tqdm(self._get_toc_items(records_list=source_records)):
            self._dedupe_pdf_toc_item(
                decision_list=decision_list,
                toc_item=toc_item,
                records=records,
                source_records=source_records,
            )

        return decision_list

    def _pdf_source_selected(self) -> bool:
        pdf_source = False
        relevant_source = [
            s
            for s in self.review_manager.settings.sources
            if str(s.filename) == self.settings.selected_source
        ]
        if len(relevant_source) > 0:
            pdf_source = "colrev.files_dir" == relevant_source[0].endpoint
        return pdf_source

    def _first_source_selected(self) -> bool:
        return (
            self.settings.selected_source
            == self.review_manager.settings.dedupe.dedupe_package_endpoints[0][
                "selected_source"
            ]
        )

    def run_dedupe(self) -> None:
        """Run the curation dedupe procedure"""

        self.dedupe_operation.merge_based_on_global_ids(apply=True)

        records = self.review_manager.dataset.load_records_dict()
        records = self._prep_records(records=records)

        # first_source should be the highest quality source
        # (which moves to md_processed first)
        first_source = self._first_source_selected()

        self._warn_on_missing_sources(first_source=first_source)

        if first_source:
            self._add_first_source_if_deduplicated(records=records)
            return

        self.review_manager.logger.info(
            "Identify duplicates between "
            f"curated_records and {self.settings.selected_source} (within toc_items)"
        )

        decision_list: list[list] = []
        # decision_list =[['ID1', 'ID2'], ...]
        if not self._pdf_source_selected():
            decision_list = self._dedupe_source(records=records)
        else:
            decision_list = self._dedupe_pdf_source(records=records)

        # Note : dedupe.apply_merges reloads the records and
        # thereby discards previous changes
        if len(decision_list) == 0:
            self.review_manager.logger.info(
                f"{Colors.GREEN}No merge-candidates identified between sets{Colors.END}"
            )
            return

        self.review_manager.logger.info(
            f"{Colors.GREEN}Duplicates identified{Colors.END}"
        )

        preferred_masterdata_sources = [
            s
            for s in self.review_manager.settings.sources
            if s.endpoint != "colrev.files_dir"
        ]
        self.dedupe_operation.apply_merges(
            id_sets=decision_list,
            preferred_masterdata_sources=preferred_masterdata_sources,
        )
        self.review_manager.dataset.create_commit(
            msg="Merge duplicate records",
        )



#! /usr/bin/env python
"""Creation of screenshots (PDFs) for online ENTRYTYPES"""
from __future__ import annotations

import time
from datetime import datetime
from pathlib import Path

import docker
import requests
import zope.interface
from docker.errors import DockerException
from pydantic import Field

import colrev.env.docker_manager
import colrev.exceptions as colrev_exceptions
import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import Fields
from colrev.constants import RecordState

# pylint: disable=duplicate-code
# pylint: disable=too-few-public-methods


@zope.interface.implementer(colrev.package_manager.interfaces.PDFGetInterface)
class WebsiteScreenshot:
    """Get PDFs from website screenshot (for "online" ENTRYTYPES)"""

    settings_class = colrev.package_manager.package_settings.DefaultSettings
    ci_supported: bool = Field(default=False)
    CHROME_BROWSERLESS_IMAGE = "browserless/chrome:latest"

    def __init__(
        self,
        *,
        pdf_get_operation: colrev.ops.pdf_get.PDFGet,
        settings: dict,
    ) -> None:
        self.settings = self.settings_class(**settings)
        self.review_manager = pdf_get_operation.review_manager
        self.pdf_get_operation = pdf_get_operation
        colrev.env.docker_manager.DockerManager.build_docker_image(
            imagename=self.CHROME_BROWSERLESS_IMAGE
        )
        pdf_get_operation.docker_images_to_stop.append(self.CHROME_BROWSERLESS_IMAGE)

    def _start_screenshot_service(self) -> None:
        """Start the screenshot service"""

        # pylint: disable=duplicate-code

        if self.screenshot_service_available():
            return

        self.review_manager.environment_manager.register_ports(["3000"])

        try:
            client = docker.from_env()

            running_containers = [
                str(container.image) for container in client.containers.list()
            ]
            if self.CHROME_BROWSERLESS_IMAGE not in running_containers:
                client.containers.run(
                    self.CHROME_BROWSERLESS_IMAGE,
                    ports={"3000/tcp": ("127.0.0.1", 3000)},
                    auto_remove=True,
                    detach=True,
                )
        except DockerException as exc:
            raise colrev_exceptions.ServiceNotAvailableException(
                f"Docker service not available ({exc}). Please install/start Docker."
            ) from exc

        i = 0
        while i < 45:
            if self.screenshot_service_available():
                break
            time.sleep(1)
            i += 1
        return

    def screenshot_service_available(self) -> bool:
        """Check if the screenshot service is available"""

        content_type_header = {"Content-type": "text/plain"}

        browserless_chrome_available = False
        try:
            ret = requests.get(
                "http://127.0.0.1:3000/",
                headers=content_type_header,
                timeout=30,
            )
            browserless_chrome_available = ret.status_code == 200

        except requests.exceptions.ConnectionError:
            pass

        if browserless_chrome_available:
            return True
        return False

    def _add_screenshot(
        self, *, record: colrev.record.record.Record, pdf_filepath: Path
    ) -> colrev.record.record.Record:
        """Add a PDF screenshot to the record"""

        if Fields.URL not in record.data:
            return record

        urldate = datetime.today().strftime("%Y-%m-%d")

        json_val = {
            Fields.URL: record.data[Fields.URL],
            "options": {
                "displayHeaderFooter": True,
                "printBackground": False,
                "format": "A2",
            },
        }

        ret = requests.post("http://127.0.0.1:3000/pdf", json=json_val, timeout=30)

        if 200 == ret.status_code:
            with open(pdf_filepath, "wb") as file:
                file.write(ret.content)

            record.update_field(
                key=Fields.FILE,
                value=str(pdf_filepath),
                source="chrome (browserless) screenshot",
            )
            # pylint: disable=colrev-direct-status-assign
            record.data.update(colrev_status=RecordState.rev_prescreen_included)
            record.update_field(
                key="urldate", value=urldate, source="chrome (browserless) screenshot"
            )

        else:
            print(
                "URL screenshot retrieval error "
                f"{ret.status_code}/{record.data['url']}"
            )

        return record

    def get_pdf(
        self, record: colrev.record.record.Record
    ) -> colrev.record.record.Record:
        """Get a PDF of the website (screenshot)"""

        if record.data[Fields.ENTRYTYPE] != "online":
            return record

        self._start_screenshot_service()

        pdf_filepath = self.review_manager.paths.pdf / Path(f"{record.data['ID']}.pdf")
        record = self._add_screenshot(record=record, pdf_filepath=pdf_filepath)

        if Fields.FILE in record.data:
            self.pdf_get_operation.import_pdf(record)

        return record
